import copy
import json
import logging
from typing import Any, Dict, List, Optional, Tuple, Type

from pydantic import BaseModel, Field, HttpUrl
from pydantic.fields import Undefined

from pait import field as pait_field
from pait.api_doc.base_parse import PaitBaseParse
from pait.g import config
from pait.model.core import PaitCoreModel
from pait.model.response import PaitBaseResponseModel
from pait.model.status import PaitStatus
from pait.util import create_pydantic_model, get_model_global_name, pait_model_schema

from .base_parse import FieldDictType, FieldSchemaTypeDict

__all__ = ["PaitOpenApi"]


class _OpenApiInfoModel(BaseModel):
    """open api info column model"""

    class _Contact(BaseModel):
        name: str
        url: str
        email: str

    class _License(BaseModel):
        name: str
        url: str

    title: str = Field("Pait Open Api")
    description: str = Field(None)
    version: str = Field("0.0.1")
    contact: _Contact = Field(None)
    license: _License = Field(None)


class _OpenApiTagModel(BaseModel):
    """openapi tag column model"""

    class _ExternalDocs(BaseModel):
        url: HttpUrl

    name: str
    description: str = Field(None)
    externalDocs: str = Field(None)


class _OpenApiServerModel(BaseModel):
    """openapi server column model"""

    url: str = Field("http://127.0.0.1")
    description: str = Field(None)


class PaitOpenApi(PaitBaseParse):
    def __init__(
        self,
        pait_dict: Dict[str, PaitCoreModel],
        title: Optional[str] = None,
        open_api_info: Optional[Dict[str, Any]] = None,
        open_api_tag_list: Optional[List[Dict[str, Any]]] = None,
        open_api_server_list: Optional[List[Dict[str, Any]]] = None,
        # default_response: Optional[...] = None,  # TODO
        type_: str = "json",
    ):
        super().__init__(pait_dict)
        self._header_keyword_dict: Dict[str, str] = {
            "Content-Type": "requestBody.content.<media-type>",
            "Accept": "responses.<code>.content.<media-type>",
            "Authorization": " security",
        }

        if not open_api_info:
            open_api_info = _OpenApiInfoModel(title=title).dict(exclude_none=True)
        else:
            _OpenApiInfoModel(**open_api_info)

        if not open_api_server_list:
            open_api_server_list = [_OpenApiServerModel().dict(exclude_none=True)]
        else:
            temp_open_api_server_list: List[Dict[str, Any]] = []
            for open_api_server in open_api_server_list:
                _OpenApiServerModel(**open_api_server)
                temp_open_api_server_list.append(open_api_server)

            open_api_server_list = temp_open_api_server_list

        if not open_api_tag_list:
            open_api_tag_list = []
        else:
            temp_open_api_tag_list: List[Dict[str, Any]] = []
            for open_api_tag in open_api_tag_list:
                _OpenApiTagModel(**open_api_tag)
                temp_open_api_tag_list.append(open_api_tag)
            open_api_tag_list = temp_open_api_tag_list
        for tag, desc in config.tag_dict.items():
            open_api_tag_list.append({"name": tag, "description": desc})

        self.open_api_dict: Dict[str, Any] = {
            "openapi": "3.0.0",
            "info": open_api_info,
            "servers": open_api_server_list,
            "tags": open_api_tag_list,
            "paths": {},
            "components": {"schemas": {}},
            # TODO
            # "security": {},
            # "externalDocs": {}
        }
        self.parse_data_2_openapi()
        if type_ == "json":
            self.content = json.dumps(self.open_api_dict, cls=config.json_encoder)
            self._content_type = ".json"
        elif type_ == "yaml":
            import yaml  # type: ignore

            self.content = yaml.dump(self.open_api_dict, sort_keys=False)
            self._content_type = ".yaml"

    def _replace_pydantic_definitions(self, schema: dict, parent_schema: Optional[dict] = None) -> None:
        """update schemas'definitions to components schemas"""
        if not parent_schema:
            parent_schema = schema
        for key, value in schema.items():
            if key == "$ref" and not value.startswith("#/components"):
                index: int = value.rfind("/") + 1
                model_key: str = value[index:]
                schema[key] = f"#/components/schemas/{model_key}"
                self.open_api_dict["components"]["schemas"][model_key] = parent_schema["definitions"][model_key]
            elif isinstance(value, dict):
                self._replace_pydantic_definitions(value, parent_schema)
            elif isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        self._replace_pydantic_definitions(item, parent_schema)

    @staticmethod
    def _snake_name_to_hump_name(name: str) -> str:
        return "".join([item.capitalize() for item in name.split("_")])

    def _pait_model_2_response(
        self,
        pait_model: PaitCoreModel,
        openapi_method_dict: dict,
    ) -> None:
        openapi_response_dict: dict = openapi_method_dict.setdefault("responses", {})
        response_schema_dict: Dict[tuple, List[Dict[str, str]]] = {}
        for resp_model_class in pait_model.response_model_list:
            resp_model: PaitBaseResponseModel = resp_model_class()
            global_model_name: str = ""
            if isinstance(resp_model.response_data, type) and issubclass(resp_model.response_data, BaseModel):
                global_model_name = get_model_global_name(resp_model.response_data)

                schema_dict: dict = copy.deepcopy(pait_model_schema(resp_model.response_data))
                self._replace_pydantic_definitions(schema_dict)
                if "definitions" in schema_dict:
                    # fix del schema dict
                    del schema_dict["definitions"]
                self.open_api_dict["components"]["schemas"].update({global_model_name: schema_dict})

            for _status_code in resp_model.status_code:
                key: tuple = (_status_code, resp_model.media_type)
                if _status_code in openapi_response_dict:
                    if resp_model.description:
                        openapi_response_dict[_status_code]["description"] += f"|{resp_model.description}"
                    if resp_model.header:
                        openapi_response_dict[_status_code]["headers"].update(resp_model.header)
                else:
                    openapi_response_dict[_status_code] = {
                        "description": resp_model.description or "",
                        "headers": resp_model.header,
                    }
                if _status_code == 204:
                    # 204 No Content, have no body.
                    # To indicate the response body is empty, do not specify a content for the response
                    continue

                openapi_response_dict[_status_code]["content"] = {}
                if resp_model.links_schema_dict:
                    openapi_response_dict[_status_code]["links"] = resp_model.links_schema_dict

                if global_model_name:
                    openapi_schema_dict: dict = {"$ref": f"#/components/schemas/{global_model_name}"}
                    if key in response_schema_dict:
                        response_schema_dict[key].append(openapi_schema_dict)
                    else:
                        response_schema_dict[key] = [openapi_schema_dict]
                elif resp_model.openapi_schema:
                    if resp_model.media_type in openapi_response_dict[_status_code]["content"]:
                        raise ValueError(
                            f"{resp_model.media_type} already exists, "
                            f"Please check {pait_model.operation_id}'s "
                            f"response model list:{pait_model.response_model_list}"
                        )
                    openapi_response_dict[_status_code]["content"][resp_model.media_type] = resp_model.openapi_schema
                else:
                    logging.warning(
                        f"Can not found response schema from {pait_model.operation_id}'s response model:{resp_model}"
                    )

        # mutli response support
        # only response example see https://swagger.io/docs/specification/describing-responses/   FAQ
        for key_tuple, path_list in response_schema_dict.items():
            status_code, media_type = key_tuple
            if len(path_list) == 1:
                openapi_schema_dict = path_list[0]
            else:
                openapi_schema_dict = {"oneOf": path_list}
            openapi_response_dict[status_code]["content"] = {media_type: {"schema": openapi_schema_dict}}

    @staticmethod
    def _field_list_2_request_parameter(
        field_class: Type[pait_field.BaseField],
        openapi_method_dict: dict,
        field_dict_list: List[FieldSchemaTypeDict],
    ) -> None:
        openapi_parameters_list: list = openapi_method_dict.setdefault("parameters", [])
        for field_dict in field_dict_list:
            param_name: str = field_dict["raw"]["param_name"]
            required: bool = field_dict["default"] is Undefined
            # openapi description must not null, but Field().description value is None
            description: str = field_dict["description"] or ""
            if field_class == pait_field.Header:
                pass
                # I don't know why the swagger doc tells me to do a title name conversion
                #  when I don't actually need one
                # See: Header Parameters in
                #   https://swagger.io/docs/specification/describing-parameters/

                # param_name = self._header_keyword_dict.get(param_name, param_name)
            elif field_class == pait_field.Path and not required:
                raise ValueError("That path parameters must have required: true, " "because they are always required")
            elif field_class == pait_field.Cookie:
                if field_dict["raw"]["field"].raw_return:
                    # fix swagger ui cookie type when pait_field.raw_return is True
                    field_dict["raw"]["schema"]["type"] = "string"
                description += (
                    " "
                    "\n"
                    ">Note for Swagger UI and Swagger Editor users: "
                    " "
                    "\n"
                    ">Cookie authentication is"
                    'currently not supported for "try it out" requests due to browser security'
                    "restrictions. "
                    "See [this issue](https://github.com/swagger-api/swagger-js/issues/1163)"
                    "for more information. "
                    "[SwaggerHub](https://swagger.io/tools/swaggerhub/)"
                    "does not have this limitation. "
                )

            # When the required parameter is False and schema.default is empty,
            # openapi will automatically read the value of schema.example
            openapi_parameters_list.append(
                {
                    "name": param_name,
                    "in": field_class.get_field_name(),
                    "required": required,
                    "description": description,
                    "schema": field_dict["raw"]["schema"],
                }
            )

    @staticmethod
    def _field_list_2_file_upload(
        field_class: Type[pait_field.BaseField],
        openapi_method_dict: dict,
        field_dict_list: List[FieldSchemaTypeDict],
    ) -> None:
        """https://swagger.io/docs/specification/describing-request-body/file-upload/"""
        openapi_request_body_dict: dict = openapi_method_dict.setdefault("requestBody", {"content": {}})
        required_column_list: List[str] = [
            field_dict["raw"]["param_name"] for field_dict in field_dict_list if field_dict["default"] is Undefined
        ]
        properties_dict: dict = {
            field_dict["raw"]["param_name"]: {
                "title": field_dict["raw"]["param_name"].capitalize(),
                "type": "string",
                "format": "binary",
            }
            for field_dict in field_dict_list
        }
        if not openapi_request_body_dict.get("required", False):
            if required_column_list:
                openapi_request_body_dict["required"] = True
        if field_class.media_type not in openapi_request_body_dict["content"]:
            openapi_request_body_dict["content"][field_class.media_type] = {
                "schema": {"type": "object", "properties": properties_dict, "required": required_column_list}
            }
        else:
            openapi_request_body_dict["content"][field_class.media_type]["schema"]["properties"].update(properties_dict)
            openapi_request_body_dict["content"][field_class.media_type]["schema"]["required"].extend(
                required_column_list
            )

    def _field_list_2_request_body(
        self,
        field_class: Type[pait_field.BaseField],
        openapi_method_dict: dict,
        field_dict_list: List[FieldSchemaTypeDict],
        operation_id: str,
    ) -> None:
        """
        gen request body schema and update request body schemas'definitions to components schemas
        Doc: https://swagger.io/docs/specification/describing-request-body/
        """
        openapi_request_body_dict: dict = openapi_method_dict.setdefault("requestBody", {"content": {}})
        annotation_dict: Dict[str, Tuple[Type, pait_field.BaseField]] = {
            field_dict["raw"]["param_name"]: (field_dict["raw"]["annotation"], field_dict["raw"]["field"])
            for field_dict in field_dict_list
        }
        _pydantic_model: Type[BaseModel] = create_pydantic_model(
            annotation_dict, class_name=f"{self._snake_name_to_hump_name(operation_id)}DynamicModel"
        )
        schema_dict = copy.deepcopy(pait_model_schema(_pydantic_model))
        # pait will disassemble and reassemble the BaseModel, so there is no way to reuse the model in openapi.
        # TODO support model
        # self._replace_pydantic_definitions(schema_dict)
        if "definitions" in schema_dict:
            del schema_dict["definitions"]
        if field_class.media_type in openapi_request_body_dict["content"]:
            for key, value in openapi_request_body_dict["content"][field_class.media_type]["schema"].items():
                if isinstance(value, list):
                    value.extend(schema_dict[key])
                elif isinstance(value, dict):
                    value.update(schema_dict[key])
        else:
            openapi_request_body_dict["content"][field_class.media_type] = {"schema": schema_dict}

        if field_class == pait_field.MultiForm:
            # pait_field.File must precede pait_field.MultiForm
            form_encoding_dict = openapi_request_body_dict["content"][field_class.media_type].setdefault("encoding", {})
            for field_dict in field_dict_list:
                if pait_field.File.media_type in openapi_request_body_dict["content"]:
                    openapi_request_body_dict["content"][field_class.media_type]["schema"]["properties"][
                        field_dict["raw"]["param_name"]
                    ]["description"] += (
                        " " " " "\n" f">Swagger UI could not support, when media_type is {pait_field.File.media_type}"
                    )
                form_encoding_dict[field_dict["raw"]["param_name"]] = field_class.openapi_serialization
            # TODO support payload?
            # https://swagger.io/docs/specification/describing-request-body/

    # flake8: noqa: C901
    def parse_data_2_openapi(self) -> None:
        for group, pait_model_list in self._group_pait_dict.items():
            for pait_model in pait_model_list:
                path: str = pait_model.openapi_path
                openapi_path_dict: dict = self.open_api_dict["paths"].setdefault(path, {})
                for method in pait_model.method_list:
                    # Documentation for each route
                    openapi_method_dict: dict = openapi_path_dict.setdefault(method.lower(), {})
                    openapi_method_dict["group"] = group  # Not an openapi standard parameter
                    if pait_model.tag:
                        openapi_method_dict["tags"] = list(pait_model.tag)
                        # Additional labeling instructions
                        for tag in pait_model.tag:
                            if tag not in {tag_dict["name"] for tag_dict in self.open_api_dict["tags"]}:
                                logging.warning(
                                    f"Can not found tag:{tag} description, set default description, "
                                    f"you can use pait.model.tag.Tag({tag}, desc='')"
                                )
                                self.open_api_dict["tags"].append(
                                    {
                                        "name": tag,
                                        "description": "",
                                    }
                                )
                    if pait_model.status in (
                        PaitStatus.abnormal,
                        PaitStatus.maintenance,
                        PaitStatus.archive,
                        PaitStatus.abandoned,
                    ):
                        openapi_method_dict["deprecated"] = True
                    openapi_method_dict["summary"] = pait_model.summary
                    openapi_method_dict["description"] = pait_model.desc
                    openapi_method_dict["operationId"] = f"{method}.{pait_model.operation_id}"

                    all_field_dict: FieldDictType = self._parse_pait_model_to_field_dict(pait_model)
                    all_field_class_list: List[Type[pait_field.BaseField]] = sorted(
                        [i for i in all_field_dict.keys()], key=lambda x: x.get_field_name()
                    )
                    for field_class in all_field_class_list:
                        field_dict_list = all_field_dict[field_class]
                        if field_class in (
                            pait_field.Cookie,
                            pait_field.Header,
                            pait_field.Path,
                            pait_field.Query,
                            pait_field.MultiQuery,
                        ):
                            self._field_list_2_request_parameter(field_class, openapi_method_dict, field_dict_list)
                        elif field_class in (pait_field.Body, pait_field.Form, pait_field.MultiForm):
                            self._field_list_2_request_body(
                                field_class, openapi_method_dict, field_dict_list, pait_model.operation_id
                            )
                        elif field_class in (pait_field.File,):
                            self._field_list_2_file_upload(field_class, openapi_method_dict, field_dict_list)
                        else:
                            logging.warning(f"Pait not support field:{field_class}")

                    if pait_model.response_model_list:
                        self._pait_model_2_response(pait_model, openapi_method_dict)



import inspect
import logging
from typing import Any, Optional, Type

from pydantic import BaseModel

from pait.exceptions import TipException
from pait.field import BaseField
from pait.util import FuncSig


def gen_tip_exc(_object: Any, exception: "Exception", parameter: Optional[inspect.Parameter] = None) -> Exception:
    """Help users understand which parameter is wrong"""
    if not parameter and (getattr(exception, "_is_tip_exc", None) or isinstance(exception, TipException)):
        return exception
    if parameter:
        param_value: BaseField = parameter.default
        annotation: Type[BaseModel] = parameter.annotation
        param_name: str = parameter.name

        parameter_value_name: str = param_value.__class__.__name__
        if param_value is parameter.empty:
            param_str: str = f"{param_name}: {annotation}"
        else:
            param_str = f"{param_name}: {annotation} = {parameter_value_name}"
            if isinstance(param_value, BaseField):
                param_str += f"(alias={param_value.alias}, default={param_value.default})"
    else:
        param_str = ""

    file: Optional[str] = None
    if isinstance(_object, FuncSig):
        _object = _object.func
    if inspect.isfunction(_object):
        title: str = "def"
        if inspect.iscoroutinefunction(_object):
            title = "async def"
        file = inspect.getfile(_object)
        line: int = inspect.getsourcelines(_object)[1]
        error_object_name: str = _object.__name__
        logging.debug(
            f"""
{title} {error_object_name}(
    ...
    {param_str} <-- error
    ...
):
    pass
"""
        )
    else:
        module: Any = inspect.getmodule(_object)
        if module:
            file = module.__file__
        line = inspect.getsourcelines(_object.__class__)[1]
        error_object_name = _object.__class__.__name__
        if "class" in error_object_name:
            error_object_name = str(_object.__class__)  # pragma: no cover
        logging.debug(f"class: `{error_object_name}`  attributes error\n    {param_str}")

    setattr(exception, "_is_tip_exc", True)
    exc_msg: str = f'File "{file}",' f" line {line}," f" in {error_object_name}." f"\nerror:{str(exception)}"
    return TipException(exc_msg, exception)



from typing import Any, Dict, List

from pait.exceptions import CheckValueError
from pait.plugin.base import BaseAsyncPlugin, BasePlugin, PluginProtocol
from pait.util import gen_tip_exc

__all__ = ["RequiredPlugin", "AsyncRequiredPlugin"]


class RequiredPluginProtocol(PluginProtocol):
    """
    Check dependencies between parameters
    """

    is_pre_core: bool = False

    def __init__(self, *, required_dict: Dict[str, List[str]]):
        super().__init__()
        self.required_dict: Dict[str, List[str]] = required_dict

    def check_param(self, *args: Any, **kwargs: Any) -> None:
        try:
            for pre_param, param_list in self.required_dict.items():
                if pre_param not in kwargs or not kwargs[pre_param]:
                    continue
                for param in param_list:
                    if kwargs.get(param, None) is None:
                        raise CheckValueError(
                            f"{pre_param} requires param {' and '.join(param_list)}, which if not none"
                        )
        except Exception as e:
            raise e from gen_tip_exc(self.pait_core_model.func, e)


class RequiredPlugin(RequiredPluginProtocol, BasePlugin):
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        self.check_param(*args, **kwargs)
        return self.call_next(*args, **kwargs)


class AsyncRequiredPlugin(RequiredPluginProtocol, BaseAsyncPlugin):
    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        self.check_param(*args, **kwargs)
        return await self.call_next(*args, **kwargs)



from tempfile import NamedTemporaryFile
from typing import IO, Any, AsyncContextManager

import aiofiles  # type: ignore
from starlette.background import BackgroundTask
from starlette.responses import FileResponse, HTMLResponse, JSONResponse, PlainTextResponse, Response

from pait.model import response
from pait.plugin.base_mock_response import BaseAsyncMockPlugin, BaseMockPlugin


class MockAsyncPlugin(BaseAsyncMockPlugin):
    def mock_response(self) -> Any:
        async def make_mock_response() -> Response:
            if issubclass(self.pait_response_model, response.PaitJsonResponseModel):
                resp: Response = JSONResponse(self.pait_response_model.get_example_value())
            elif issubclass(self.pait_response_model, response.PaitTextResponseModel):
                resp = PlainTextResponse(
                    self.pait_response_model.get_example_value(), media_type=self.pait_response_model.media_type
                )
            elif issubclass(self.pait_response_model, response.PaitHtmlResponseModel):
                resp = HTMLResponse(
                    self.pait_response_model.get_example_value(), media_type=self.pait_response_model.media_type
                )
            elif issubclass(self.pait_response_model, response.PaitFileResponseModel):
                named_temporary_file: AsyncContextManager = aiofiles.tempfile.NamedTemporaryFile()  # type: ignore
                f: Any = await named_temporary_file.__aenter__()
                await f.write(self.pait_response_model.get_example_value())
                await f.seek(0)

                async def close_file() -> None:
                    await named_temporary_file.__aexit__(None, None, None)

                resp = FileResponse(
                    f.name, media_type=self.pait_response_model.media_type, background=BackgroundTask(close_file)
                )
            else:
                raise NotImplementedError(f"make_mock_response not support {self.pait_response_model}")
            resp.status_code = self.pait_response_model.status_code[0]
            if self.pait_response_model.header:
                resp.headers.update(self.pait_response_model.header)
            return resp

        return make_mock_response()


class MockPlugin(BaseMockPlugin):
    def mock_response(self) -> Any:
        if issubclass(self.pait_response_model, response.PaitJsonResponseModel):
            resp: Response = JSONResponse(self.pait_response_model.get_example_value())
        elif issubclass(self.pait_response_model, response.PaitTextResponseModel):
            resp = PlainTextResponse(
                self.pait_response_model.get_example_value(), media_type=self.pait_response_model.media_type
            )
        elif issubclass(self.pait_response_model, response.PaitHtmlResponseModel):
            resp = HTMLResponse(
                self.pait_response_model.get_example_value(), media_type=self.pait_response_model.media_type
            )
        elif issubclass(self.pait_response_model, response.PaitFileResponseModel):
            named_temporary_file: IO = NamedTemporaryFile(delete=True)
            f: Any = named_temporary_file.__enter__()
            f.write(self.pait_response_model.get_example_value())
            f.seek(0)

            def close_file() -> None:
                named_temporary_file.__exit__(None, None, None)

            resp = FileResponse(
                f.name, media_type=self.pait_response_model.media_type, background=BackgroundTask(close_file)
            )
        else:
            raise NotImplementedError(f"make_mock_response not support {self.pait_response_model}")
        resp.status_code = self.pait_response_model.status_code[0]
        if self.pait_response_model.header:
            resp.headers.update(self.pait_response_model.header)
        return resp



import inspect
import json
import sys
from datetime import datetime
from decimal import Decimal
from enum import Enum
from json import JSONEncoder
from typing import TYPE_CHECKING, Any, Dict, ForwardRef, List, Optional, Type, Union, get_type_hints

from pydantic import BaseModel
from pydantic.fields import Undefined, UndefinedType

from pait.field import BaseField, Depends, is_pait_field

from ._types import ParseTypeError, parse_typing

if TYPE_CHECKING:
    from pait.model.response import PaitBaseResponseModel


json_type_default_value_dict: Dict[str, Any] = {
    "null": None,
    "bool": True,
    "boolean": True,
    "string": "",
    "number": 0.0,
    "float": 0.0,
    "integer": 0,
    "object": {},
    "array": [],
}

python_type_default_value_dict: Dict[type, Any] = {
    bool: True,
    float: 0.0,
    int: 0,
    str: "",
    list: [],
    tuple: (),
    dict: {},
    datetime: datetime.fromtimestamp(0),
    Decimal: Decimal("0.0"),
}


def get_real_annotation(annotation: Union[Type, str], target_obj: Any) -> Type:
    if not isinstance(annotation, str):
        return annotation
    if inspect.isclass(target_obj) and annotation in target_obj.__dict__:
        new_annotation: Type = target_obj.__dict__[annotation]
    else:
        # get real type
        value: ForwardRef = ForwardRef(annotation, is_argument=False)
        new_annotation = value._evaluate(sys.modules[target_obj.__module__].__dict__, None)  # type: ignore
        if not new_annotation:
            raise RuntimeError(f"get real annotation from {target_obj} fail")  # pragma: no cover
    return new_annotation


def get_pydantic_annotation(key: str, pydantic_base_model: Type[BaseModel]) -> Type:
    annotation: Type = pydantic_base_model.__annotations__[key]
    annotation = get_real_annotation(annotation, pydantic_base_model)

    if getattr(annotation, "real", None) and annotation != bool:
        # support like pydantic.ConstrainedIntValue
        annotation = annotation.real.__objclass__  # type: ignore
    return annotation


def get_pait_response_model(
    response_model_list: List[Type["PaitBaseResponseModel"]],
    target_pait_response_class: Optional[Type["PaitBaseResponseModel"]] = None,
    find_core_response_model: bool = False,
) -> Type["PaitBaseResponseModel"]:
    if target_pait_response_class:
        core_response_list: List[Type["PaitBaseResponseModel"]] = [
            i for i in response_model_list if i.is_core and issubclass(i, target_pait_response_class)
        ]
    else:
        core_response_list = [i for i in response_model_list if i.is_core]
    if find_core_response_model:
        if len(core_response_list) != 1:
            raise RuntimeError("Multiple pait response models were found")
        return core_response_list[0]
    else:
        return (core_response_list or response_model_list)[0]


def gen_example_value_from_python(obj: Any) -> Any:
    if isinstance(obj, dict):
        new_dict: dict = {}
        for key, value in obj.items():
            new_dict[key] = gen_example_value_from_python(value)
        return new_dict
    else:
        return python_type_default_value_dict.get(type(obj), obj)


def gen_example_dict_from_pydantic_base_model(
    pydantic_base_model: Type[BaseModel], use_example_value: bool = True
) -> dict:
    gen_dict: Dict[str, Any] = {}
    for key, model_field in pydantic_base_model.__fields__.items():
        if use_example_value:
            example_value: Any = model_field.field_info.extra.get("example", Undefined)
            if not isinstance(example_value, UndefinedType):
                if inspect.isfunction(example_value):
                    example_value = example_value()
                elif isinstance(example_value, Enum):
                    example_value = example_value.value
                gen_dict[key] = example_value
                continue

        if not isinstance(model_field.field_info.default, UndefinedType):
            gen_dict[key] = model_field.field_info.default
        elif model_field.field_info.default_factory and not isinstance(
            model_field.field_info.default_factory, UndefinedType
        ):
            gen_dict[key] = model_field.field_info.default_factory()
        else:
            annotation: Type = get_pydantic_annotation(key, pydantic_base_model)

            if issubclass(annotation, BaseModel):
                gen_dict[key] = gen_example_dict_from_pydantic_base_model(annotation)
            else:
                try:
                    parse_typing_result: Union[List[Type[Any]], Type] = parse_typing(annotation)
                    if isinstance(parse_typing_result, list):
                        real_type: Type = parse_typing_result[0]
                    else:
                        real_type = parse_typing_result
                except ParseTypeError:
                    real_type = annotation
                if issubclass(real_type, Enum):
                    gen_dict[key] = [i for i in real_type.__members__.values()][0].value
                else:
                    gen_dict[key] = python_type_default_value_dict[real_type]
    return gen_dict


def gen_example_dict_from_schema(
    schema_dict: Dict[str, Any],
    definition_dict: Optional[dict] = None,
    use_example_value: bool = True,
) -> Dict[str, Any]:
    gen_dict: Dict[str, Any] = {}
    property_dict: Dict[str, Any] = schema_dict["properties"]
    if not definition_dict:
        _definition_dict: dict = schema_dict.get("definitions", {})
    else:
        _definition_dict = definition_dict
    for key, value in property_dict.items():
        if "items" in value and value["type"] == "array":
            if "$ref" in value["items"]:
                model_key: str = value["items"]["$ref"].split("/")[-1]
                gen_dict[key] = [
                    gen_example_dict_from_schema(
                        _definition_dict.get(model_key, {}), _definition_dict, use_example_value=use_example_value
                    )
                ]
            else:
                gen_dict[key] = []
        elif "$ref" in value:
            model_key = value["$ref"].split("/")[-1]
            gen_dict[key] = gen_example_dict_from_schema(
                _definition_dict.get(model_key, {}), _definition_dict, use_example_value=use_example_value
            )
        else:
            if "example" in value:
                gen_dict[key] = value["example"]
                if inspect.isfunction(gen_dict[key]):
                    gen_dict[key] = gen_dict[key]()
            elif "default" in value:
                gen_dict[key] = value["default"]
            else:
                if "type" in value:
                    if value["type"] not in json_type_default_value_dict:
                        raise KeyError(f"Can not found type: {key} in json type")
                    gen_dict[key] = json_type_default_value_dict[value["type"]]
                else:
                    gen_dict[key] = "object"
            if isinstance(gen_dict[key], Enum):
                gen_dict[key] = gen_dict[key].value
    return gen_dict


def gen_example_json_from_schema(schema_dict: Dict[str, Any], cls: Optional[Type[JSONEncoder]] = None) -> str:
    return json.dumps(gen_example_dict_from_schema(schema_dict), cls=cls)


def get_parameter_list_from_pydantic_basemodel(pait_model: Type[BaseModel]) -> List["inspect.Parameter"]:
    """get class parameter list by attributes, if attributes not default value, it will be set `Undefined`"""
    parameter_list: Optional[List["inspect.Parameter"]] = getattr(pait_model, "_parameter_list", None)
    if parameter_list is not None:
        return parameter_list
    parameter_list = []
    for key, model_field in pait_model.__fields__.items():
        if not is_pait_field(model_field.field_info):
            raise TypeError(f"{model_field.field_info} must instance {BaseField} or {Depends}")

        parameter = inspect.Parameter(
            key,
            inspect.Parameter.POSITIONAL_ONLY,
            default=model_field.field_info,
            annotation=get_pydantic_annotation(key, pait_model),
        )
        parameter_list.append(parameter)

    setattr(pait_model, "_parameter_list", parameter_list)
    return parameter_list


def get_parameter_list_from_class(cbv_class: Type) -> List["inspect.Parameter"]:
    """get class parameter list by attributes, if attributes not default value, it will be set `Undefined`"""
    parameter_list: Optional[List["inspect.Parameter"]] = getattr(cbv_class, "_parameter_list", None)
    if parameter_list is not None:
        return parameter_list
    parameter_list = []
    if hasattr(cbv_class, "__annotations__"):
        for param_name, param_annotation in get_type_hints(cbv_class).items():
            default: Any = getattr(cbv_class, param_name, Undefined)
            if not is_pait_field(default):
                continue
            parameter: "inspect.Parameter" = inspect.Parameter(
                param_name,
                inspect.Parameter.POSITIONAL_ONLY,
                default=default,
                annotation=param_annotation,
            )
            parameter_list.append(parameter)
    setattr(cbv_class, "_parameter_list", parameter_list)
    return parameter_list



from typing import Any, Callable, Optional

from starlette.responses import JSONResponse

from pait.plugin.check_json_resp import AsyncCheckJsonRespPlugin as _AsyncCheckJsonRespPlugin
from pait.plugin.check_json_resp import CheckJsonRespPlugin as _CheckJsonRespPlugin

__all__ = ["AsyncCheckJsonRespPlugin", "CheckJsonRespPlugin"]


class AsyncCheckJsonRespPlugin(_AsyncCheckJsonRespPlugin):
    def __init__(
        self,
        check_resp_fn: Callable,
        status_code: int = 200,
        headers: Optional[dict] = None,
        media_type: Optional[str] = None,
    ) -> None:
        super(AsyncCheckJsonRespPlugin, self).__init__(check_resp_fn=check_resp_fn)
        self.status_code: int = status_code
        self.headers: Optional[dict] = headers
        self.media_type: Optional[str] = media_type

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        response: Any = await super().__call__(*args, **kwargs)
        return JSONResponse(response, status_code=self.status_code, headers=self.headers, media_type=self.media_type)


class CheckJsonRespPlugin(_CheckJsonRespPlugin):
    def __init__(
        self,
        check_resp_fn: Callable,
        status_code: int = 200,
        headers: Optional[dict] = None,
        media_type: Optional[str] = None,
    ) -> None:
        super(CheckJsonRespPlugin, self).__init__(check_resp_fn=check_resp_fn)
        self.status_code: int = status_code
        self.headers: Optional[dict] = headers
        self.media_type: Optional[str] = media_type

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        response: Any = super().__call__(*args, **kwargs)
        return JSONResponse(response, status_code=self.status_code, headers=self.headers, media_type=self.media_type)



from typing import Any, Dict, Optional, Tuple, Type, Union

from pydantic import BaseModel

from pait.util import gen_example_dict_from_pydantic_base_model


class PaitBaseResponseModel(object):
    """response model https://swagger.io/docs/specification/describing-responses/"""

    # Used for mock response and response checking to determine if the response model is the core response model
    is_core: bool = False

    # response data
    response_data: Union[Type[BaseModel], str, bytes]
    # response media type
    media_type: str

    # response name
    name: Optional[str] = None
    # response description
    description: Optional[str] = None
    # response header
    header: dict = {}
    # response status code
    status_code: Tuple[int] = (200,)

    # The value of this response in openapi.schema
    # if value is empty, pait will auto gen response model and set to openapi.schema
    openapi_schema: Optional[dict] = None

    # links model
    links_schema_dict: Dict[str, dict] = {}

    @classmethod
    def is_base_model_response_data(cls) -> bool:
        return isinstance(cls.response_data, type) and issubclass(cls.response_data, BaseModel)

    @classmethod
    def get_example_value(cls, **extra: Any) -> Any:
        return cls.response_data

    @classmethod
    def register_link_schema(cls, link_schema: dict) -> None:
        cls.links_schema_dict.update(link_schema)


class PaitJsonResponseModel(PaitBaseResponseModel):
    response_data: Type[BaseModel]
    media_type: str = "application/json"

    @classmethod
    def get_example_value(cls, **extra: Any) -> dict:
        return gen_example_dict_from_pydantic_base_model(cls.response_data)

    @classmethod
    def get_default_dict(cls, **extra: Any) -> dict:
        return gen_example_dict_from_pydantic_base_model(cls.response_data, use_example_value=False)


class PaitResponseModel(PaitJsonResponseModel):
    """
    PaitJsonResponseModel alias
    Compatible versions below 0.7
    """


class PaitTextResponseModel(PaitBaseResponseModel):
    response_data: str = "pait example data"
    media_type: str = "text/plain"

    openapi_schema: dict = {"type": "string", "example": response_data}


class PaitHtmlResponseModel(PaitBaseResponseModel):
    response_data: str = "<h1>Pait example html</h1>"
    media_type: str = "text/html"

    openapi_schema: dict = {"type": "string", "example": response_data}


class PaitFileResponseModel(PaitBaseResponseModel):
    response_data: bytes = b"pait example bytes"
    media_type: str = "application/octet-stream"

    openapi_schema: dict = {"type": "string", "format": "binary"}



import asyncio
import copy
import inspect
import logging
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from types import TracebackType
from typing import TYPE_CHECKING, Any, Callable, Coroutine, Dict, List, Mapping, Optional, Tuple, Type, Union

from pydantic import BaseConfig, BaseModel
from pydantic.fields import Undefined, UndefinedType

from pait import field
from pait.app.base import BaseAppHelper
from pait.exceptions import (
    FieldValueTypeException,
    NotFoundFieldException,
    NotFoundValueException,
    PaitBaseException,
    ParseTypeError,
)
from pait.field import BaseField
from pait.plugin.base import BaseAsyncPlugin, BasePlugin, PluginProtocol
from pait.util import (
    FuncSig,
    create_pydantic_model,
    gen_tip_exc,
    get_func_sig,
    get_parameter_list_from_class,
    get_parameter_list_from_pydantic_basemodel,
    is_type,
)

if TYPE_CHECKING:
    from pait.model.core import PaitCoreModel


def raise_multiple_exc(exc_list: List[Exception]) -> None:
    """
    Multiple exceptions may be thrown during the parsing process, and these will be thrown one by one like a stack
    """
    if not exc_list:
        return
    try:
        raise exc_list.pop()
    finally:
        raise_multiple_exc(exc_list)


def parameter_2_basemodel(
    parameter_value_dict: Dict["inspect.Parameter", Any],
    pydantic_config: Type[BaseConfig],
    use_pydantic_base_model_alias: bool = False,
) -> BaseModel:
    """Convert all parameters into pydantic mods"""
    annotation_dict: Dict[str, Tuple[Type, Any]] = {}
    param_value_dict: Dict[str, Any] = {}
    for parameter, value in parameter_value_dict.items():
        if isinstance(parameter.default, BaseField) and parameter.default.alias:
            # Resolve the key mismatch between Field.alias and request value
            param_field: field.BaseField = copy.deepcopy(parameter.default)
            param_field.default = value
            base_model_key: str = parameter.default.alias if use_pydantic_base_model_alias else parameter.name
        else:
            param_field = parameter.default
            base_model_key = parameter.name
        annotation_dict[base_model_key] = (parameter.annotation, param_field)
        param_value_dict[base_model_key] = value

    return create_pydantic_model(annotation_dict, pydantic_config=pydantic_config)(**param_value_dict)


class ParamHandlerMixin(PluginProtocol):
    def __post_init__(self, pait_core_model: "PaitCoreModel", args: tuple, kwargs: dict) -> None:
        super(ParamHandlerMixin, self).__post_init__(pait_core_model, args, kwargs)

        # cbv handle
        self.cbv_instance: Optional[Any] = None
        self.cbv_type: Optional[Type] = None
        if self.args and self.args[0].__class__.__name__ in self.pait_core_model.func.__qualname__:
            self.cbv_instance = self.args[0]
            self.cbv_type = self.cbv_instance.__class__
        # else:
        #     cbv_type = getattr(inspect.getmodule(func), func.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0])

        self._app_helper: BaseAppHelper = pait_core_model.app_helper_class(self.cbv_instance, self.args, self.kwargs)

        self.pre_depend_list: List[Callable] = pait_core_model.pre_depend_list
        self.pydantic_model_config: Type[BaseConfig] = pait_core_model.pydantic_model_config

        if self.cbv_type:
            self.cbv_param_list: List["inspect.Parameter"] = get_parameter_list_from_class(self.cbv_type)
        else:
            self.cbv_param_list = []

    @classmethod
    def check_depend_handle(cls, func: Callable) -> Any:
        func_sig: FuncSig = get_func_sig(func)
        cls.check_param_field_handle(func_sig, func_sig.param_list)

    @staticmethod
    def check_field_type(value: Any, target_type: Any, error_msg: str) -> None:
        if isinstance(value, UndefinedType):
            return
        if inspect.isfunction(value):
            value = value()
        try:
            result = is_type(type(value), target_type)
        except ParseTypeError:
            result = isinstance(value, target_type)

        if not result:
            raise ParseTypeError(error_msg)

    @classmethod
    def check_param_field_handle(
        cls,
        _object: Union[FuncSig, Type, None],
        param_list: List["inspect.Parameter"],
    ) -> None:

        for parameter in param_list:
            try:
                if parameter.default != parameter.empty:
                    # kwargs param
                    # support model: def demo(pydantic.BaseModel: BaseModel = pait.field.BaseField())
                    if isinstance(parameter.default, field.Depends):
                        cls.check_depend_handle(parameter.default.func)
                    elif isinstance(parameter.default, field.BaseField):
                        try:
                            cls.check_field_type(
                                parameter.default.default,
                                parameter.annotation,
                                f"{parameter.name}'s Field.default value must {parameter.annotation}",
                            )
                            if parameter.default.default_factory:
                                cls.check_field_type(
                                    parameter.default.default_factory(),
                                    parameter.annotation,
                                    f"{parameter.name}'s Field.default_factory value must {parameter.annotation}",
                                )
                            cls.check_field_type(
                                parameter.default.extra.get("example", Undefined),
                                parameter.annotation,
                                f"{parameter.name}'s Field.example value must {parameter.annotation}",
                            )
                        except ParseTypeError as e:
                            raise FieldValueTypeException(parameter.name, str(e))
                else:
                    # args param
                    # support model: model: ModelType
                    if issubclass(parameter.annotation, BaseModel):
                        get_parameter_list_from_pydantic_basemodel(parameter.annotation)
            except PaitBaseException as e:
                raise gen_tip_exc(_object, e, parameter)

    @classmethod
    def cls_hook_by_core_model(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        # check param from pre depend
        for pre_depend in pait_core_model.pre_depend_list:
            cls.check_depend_handle(pre_depend)

        # check param from func
        func_sig: FuncSig = get_func_sig(pait_core_model.func)
        cls.check_param_field_handle(func_sig, func_sig.param_list)

        # TODO support cbv class Attribute
        # I don't know how to get the class of the decorated function at the initialization of the decorator,
        # which may be an unattainable requirement
        return kwargs

    def _set_parameter_value_to_args(self, parameter: inspect.Parameter, func_args: list) -> bool:
        """Extract the self parameter of the cbv handler,
        the request parameter of the route and the parameter of type PaitBaseModel,
        and check if there are any other parameters that do not meet the conditions
        """
        if self.cbv_instance and not func_args:
            # first parma must self param
            func_args.append(self._app_helper.cbv_instance)
        elif self._app_helper.check_request_type(parameter.annotation):
            # support request param(def handle(request: Request))
            func_args.append(self._app_helper.request)
        elif issubclass(parameter.annotation, BaseModel):
            return True
        else:
            logging.warning(f"Pait not support args: {parameter}")  # pragma: no cover
        return False

    def request_value_handle(
        self,
        parameter: inspect.Parameter,
        request_value: Any,
        base_model_dict: Optional[Dict[str, Any]],
        parameter_value_dict: Dict["inspect.Parameter", Any],
    ) -> None:
        """parse request_value and set to base_model_dict or parameter_value_dict"""
        param_value: BaseField = parameter.default
        annotation: Type[BaseModel] = parameter.annotation
        param_name: str = parameter.name

        if not isinstance(param_value, BaseField):
            # not support
            # raise PaitBaseException(f"must use {BaseField.__class__.__name__}, no {param_value}")
            return  # pragma: no cover
        elif (
            isinstance(request_value, Mapping)
            # some type like dict, but not isinstance Mapping, e.g: werkzeug.datastructures.EnvironHeaders
            or self._app_helper.check_header_type(type(request_value))
            or self._app_helper.check_form_type(type(request_value))
        ):
            if not param_value.raw_return:
                # The code execution effect should be consistent with the generated documentation, no diversity
                # so remove code:
                #   > if type(param_value.alias) is str and param_value.alias in request_value:
                if type(param_value.alias) is str and param_value.alias:
                    request_value_key: str = param_value.alias
                else:
                    request_value_key = param_name
                request_value = request_value.get(request_value_key, param_value.default)
                if isinstance(request_value, UndefinedType):
                    if param_value.default_factory:
                        request_value = param_value.default_factory()
                    else:
                        raise NotFoundValueException(parameter.name, f"Can not found {parameter.name} value")

            if base_model_dict is not None and inspect.isclass(annotation) and issubclass(annotation, BaseModel):
                # parse annotation is pydantic.BaseModel and base_model_dict not None
                base_model_dict[parameter.name] = annotation(**request_value)
            else:
                # parse annotation is python type and pydantic.field
                parameter_value_dict[parameter] = request_value
        else:
            parameter_value_dict[parameter] = request_value

    def get_request_value_from_parameter(self, parameter: inspect.Parameter) -> Union[Any, Coroutine]:
        assert isinstance(parameter.default, BaseField), f"{parameter.name}'s value must pait field"
        field_name: str = parameter.default.get_field_name()
        # Note: not use hasattr with LazyProperty (
        #   because hasattr will calling getattr(obj, name) and catching AttributeError,
        # )
        app_field_func: Optional[Callable] = getattr(self._app_helper, field_name, None)
        if app_field_func is None:
            raise NotFoundFieldException(
                parameter.name, f"field: {field_name} not found in {self._app_helper}"
            )  # pragma: no cover
        return app_field_func()


class ParamHandler(ParamHandlerMixin, BasePlugin):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._contextmanager_list: List[AbstractContextManager] = []

    def param_handle(
        self,
        _object: Union[FuncSig, Type, None],
        param_list: List["inspect.Parameter"],
        use_pydantic_base_model_alias: bool = False,
    ) -> Tuple[List[Any], Dict[str, Any]]:
        args_param_list: List[Any] = []
        kwargs_param_dict: Dict[str, Any] = {}

        single_field_dict: Dict["inspect.Parameter", Any] = {}

        for parameter in param_list:
            try:
                if parameter.default != parameter.empty:
                    # kwargs param
                    # support model: def demo(pydantic.BaseModel: BaseModel = pait.field.BaseField())
                    if isinstance(parameter.default, field.Depends):
                        kwargs_param_dict[parameter.name] = self._depend_handle(parameter.default.func)
                    else:
                        request_value: Any = self.get_request_value_from_parameter(parameter)
                        self.request_value_handle(
                            parameter,
                            request_value,
                            kwargs_param_dict,
                            single_field_dict,
                        )
                else:
                    # args param
                    # support model: model: ModelType
                    self.set_parameter_value_to_args(parameter, args_param_list)
            except PaitBaseException as e:
                raise gen_tip_exc(_object, e, parameter)
        # support field: def demo(demo_param: int = pait.field.BaseField())
        if single_field_dict:
            try:
                kwargs_param_dict.update(
                    parameter_2_basemodel(
                        single_field_dict,
                        self.pydantic_model_config,
                        use_pydantic_base_model_alias=use_pydantic_base_model_alias,
                    ).dict(),
                )
            except Exception as e:
                raise e from gen_tip_exc(_object, e)
        return args_param_list, kwargs_param_dict

    def set_parameter_value_to_args(self, parameter: inspect.Parameter, func_args: list) -> None:
        """use func_args param faster return and extend func_args"""
        if not self._set_parameter_value_to_args(parameter, func_args):
            return
        # support pait_model param(def handle(model: PaitBaseModel))
        _pait_model: Type[BaseModel] = parameter.annotation
        _, kwargs = self.param_handle(
            None, get_parameter_list_from_pydantic_basemodel(_pait_model), use_pydantic_base_model_alias=True
        )
        func_args.append(_pait_model(**kwargs))

    def _depend_handle(self, func: Callable) -> Any:
        func_sig: FuncSig = get_func_sig(func)
        _func_args, _func_kwargs = self.param_handle(func_sig, func_sig.param_list)
        func_result: Any = func(*_func_args, **_func_kwargs)
        if isinstance(func_result, AbstractContextManager):
            self._contextmanager_list.append(func_result)
            return func_result.__enter__()
        else:
            return func_result

    def _gen_param(self) -> None:
        # check param from pre depend
        for pre_depend in self.pre_depend_list:
            self._depend_handle(pre_depend)

        # gen and check param from func
        func_sig: FuncSig = get_func_sig(self.pait_core_model.func)
        self.args, self.kwargs = self.param_handle(func_sig, func_sig.param_list)

        # gen and check param from class
        if self.cbv_param_list and self.cbv_type:
            _, kwargs = self.param_handle(self.cbv_type, self.cbv_param_list)
            self.cbv_instance.__dict__.update(kwargs)
        return None

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        with self:
            return self.call_next(*self.args, **self.kwargs)

    def __enter__(self) -> "ParamHandler":
        try:
            self._gen_param()
            return self
        except Exception as e:
            raise e from gen_tip_exc(self.call_next, e)

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> Optional[bool]:
        exc_list: List[Exception] = []
        for contextmanager in self._contextmanager_list:
            try:
                contextmanager.__exit__(exc_type, exc_val, exc_tb)
            except Exception as e:
                exc_list.append(e)
            else:
                if exc_type and issubclass(exc_type, Exception):
                    exc_list.append(exc_type(exc_val).with_traceback(exc_tb))
        if exc_list:
            raise_multiple_exc(exc_list)
            return True
        else:
            return False


class AsyncParamHandler(ParamHandlerMixin, BaseAsyncPlugin):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._contextmanager_list: List[Union[AbstractAsyncContextManager, AbstractContextManager]] = []

    async def param_handle(
        self,
        _object: Union[FuncSig, Type],
        param_list: List["inspect.Parameter"],
        use_pydantic_base_model_alias: bool = False,
    ) -> Tuple[List[Any], Dict[str, Any]]:
        args_param_list: List[Any] = []
        kwargs_param_dict: Dict[str, Any] = {}
        single_field_dict: Dict["inspect.Parameter", Any] = {}

        for parameter in param_list:
            try:
                if parameter.default != parameter.empty:
                    # kwargs param
                    # support like: def demo(pydantic.BaseModel: BaseModel = pait.field.BaseField())
                    if isinstance(parameter.default, field.Depends):
                        kwargs_param_dict[parameter.name] = await self._depend_handle(parameter.default.func)
                    else:
                        request_value: Any = self.get_request_value_from_parameter(parameter)
                        if asyncio.iscoroutine(request_value) or asyncio.isfuture(request_value):
                            request_value = await request_value
                        self.request_value_handle(parameter, request_value, kwargs_param_dict, single_field_dict)
                else:
                    # args param
                    # support model: model: ModelType
                    await self.set_parameter_value_to_args(_object, parameter, args_param_list)
            except PaitBaseException as e:
                raise gen_tip_exc(_object, e, parameter)
        # support field: def demo(demo_param: int = pait.field.BaseField())
        if single_field_dict:
            try:
                kwargs_param_dict.update(
                    parameter_2_basemodel(
                        single_field_dict,
                        self.pydantic_model_config,
                        use_pydantic_base_model_alias=use_pydantic_base_model_alias,
                    ).dict()
                )
            except Exception as e:
                raise e from gen_tip_exc(_object, e)
        return args_param_list, kwargs_param_dict

    async def set_parameter_value_to_args(
        self, _object: Union[FuncSig, Type], parameter: inspect.Parameter, func_args: list
    ) -> None:
        """use func_args param faster return and extend func_args"""
        if not self._set_parameter_value_to_args(parameter, func_args):
            return
        _pait_model: Type[BaseModel] = parameter.annotation
        _, kwargs = await self.param_handle(
            _object, get_parameter_list_from_pydantic_basemodel(_pait_model), use_pydantic_base_model_alias=True
        )
        func_args.append(_pait_model(**kwargs))

    async def _depend_handle(self, func: Callable) -> Any:
        func_sig: FuncSig = get_func_sig(func)
        _func_args, _func_kwargs = await self.param_handle(func_sig, func_sig.param_list)
        func_result: Any = func(*_func_args, **_func_kwargs)
        if asyncio.iscoroutine(func_result):
            func_result = await func_result
        if isinstance(func_result, AbstractAsyncContextManager):
            self._contextmanager_list.append(func_result)
            return await func_result.__aenter__()
        elif isinstance(func_result, AbstractContextManager):
            self._contextmanager_list.append(func_result)
            return func_result.__enter__()
        else:
            return func_result

    async def _gen_param(self) -> None:
        # check param from pre depend
        for pre_depend in self.pre_depend_list:
            await self._depend_handle(pre_depend)

        # gen and check param from func
        func_sig: FuncSig = get_func_sig(self.pait_core_model.func)
        self.args, self.kwargs = await self.param_handle(func_sig, func_sig.param_list)

        # gen and check param from class
        if self.cbv_param_list and self.cbv_type:
            _, kwargs = await self.param_handle(self.cbv_type, self.cbv_param_list)
            self.cbv_instance.__dict__.update(kwargs)
        return None

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        async with self:
            return await self.call_next(*self.args, **self.kwargs)

    async def __aenter__(self) -> "AsyncParamHandler":
        try:
            await self._gen_param()
            return self
        except Exception as e:
            raise e from gen_tip_exc(self.call_next, e)

    async def __aexit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> Optional[bool]:
        exc_list: List[Exception] = []
        for contextmanager in self._contextmanager_list:
            try:
                if isinstance(contextmanager, AbstractContextManager):
                    contextmanager.__exit__(exc_type, exc_val, exc_tb)
                else:
                    await contextmanager.__aexit__(exc_type, exc_val, exc_tb)
            except Exception as e:
                exc_list.append(e)
            else:
                if exc_type and issubclass(exc_type, Exception):
                    exc_list.append(exc_type(exc_val).with_traceback(exc_tb))
        if exc_list:
            raise_multiple_exc(exc_list)
            return True
        else:
            return False



import asyncio
from concurrent import futures
from typing import Any, Callable, Optional


class _BoundClass(object):
    pass


class LazyProperty:
    """Cache field computing resources
    >>> class Demo:
    ...     @LazyProperty()
    ...     def value(self, value):
    ...         return value * value
    """

    def __call__(self, func: Callable) -> Callable:
        key: str = f"{self.__class__.__name__}_{func.__name__}_future"
        _bound_class: _BoundClass = _BoundClass()
        if not asyncio.iscoroutinefunction(func):

            def wrapper(*args: Any, **kwargs: Any) -> Any:
                if args and args[0].__class__.__name__ in func.__qualname__:
                    class_: Any = args[0]
                else:
                    class_ = _bound_class
                future: Optional[futures.Future] = getattr(class_, key, None)
                if not future:
                    future = futures.Future()
                    result: Any = func(*args, **kwargs)
                    future.set_result(result)
                    setattr(class_, key, future)
                    return result
                return future.result()

            return wrapper
        else:

            async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
                if args and args[0].__class__.__name__ in func.__qualname__:
                    class_: Any = args[0]
                else:
                    class_ = _bound_class
                future: Optional[asyncio.Future] = getattr(class_, key, None)
                if not future:
                    future = asyncio.Future()
                    result: Any = await func(*args, **kwargs)
                    future.set_result(result)
                    setattr(class_, key, future)
                    return result
                return future.result()

            return async_wrapper



from ._func_sig import FuncSig, get_func_sig
from ._gen_tip import gen_tip_exc
from ._i18n import I18n, I18nTypedDict, change_local, i18n_config_dict, i18n_local, join_i18n
from ._lazy_property import LazyProperty
from ._pydantic_util import create_pydantic_model, get_model_global_name, pait_get_model_name_map, pait_model_schema
from ._types import is_type, parse_typing
from ._util import (
    Undefined,
    UndefinedType,
    gen_example_dict_from_pydantic_base_model,
    gen_example_dict_from_schema,
    gen_example_json_from_schema,
    gen_example_value_from_python,
    get_pait_response_model,
    get_parameter_list_from_class,
    get_parameter_list_from_pydantic_basemodel,
    get_real_annotation,
    json_type_default_value_dict,
    python_type_default_value_dict,
)



from tempfile import NamedTemporaryFile

from flask import Response, jsonify, make_response, send_from_directory

from pait.model import response
from pait.plugin.base_mock_response import BaseMockPlugin

__all__ = ["MockPlugin"]


class MockPlugin(BaseMockPlugin):
    def mock_response(self) -> Response:
        if issubclass(self.pait_response_model, response.PaitJsonResponseModel):
            resp: Response = jsonify(self.pait_response_model.get_example_value())
        elif issubclass(self.pait_response_model, response.PaitTextResponseModel) or issubclass(
            self.pait_response_model, response.PaitHtmlResponseModel
        ):
            resp = make_response(self.pait_response_model.get_example_value())
            resp.mimetype = self.pait_response_model.media_type
        elif issubclass(self.pait_response_model, response.PaitFileResponseModel):
            with NamedTemporaryFile(delete=True) as temporary_file:
                temporary_file.write(self.pait_response_model.get_example_value())
                temporary_file.seek(0)
                _, f_path, f_filename = temporary_file.name.split("/")
                resp = send_from_directory("/" + f_path, f_filename, mimetype=self.pait_response_model.media_type)
        else:
            raise NotImplementedError(f"make_mock_response not support {self.pait_response_model}")
        resp.status_code = self.pait_response_model.status_code[0]
        if self.pait_response_model.header:
            resp.headers.update(self.pait_response_model.header)  # type: ignore
        return resp



from typing import Callable, Dict, Optional

from tornado.web import Application

from pait.g import pait_data
from pait.model.core import PaitCoreModel

from ._app_helper import AppHelper

__all__ = ["load_app"]


def load_app(app: Application, project_name: str = "") -> Dict[str, PaitCoreModel]:
    """Read data from the route that has been registered to `pait`"""
    _pait_data: Dict[str, PaitCoreModel] = {}
    for rule in app.wildcard_router.rules:
        path: str = rule.matcher.regex.pattern  # type: ignore
        openapi_path: str = path
        # replace path regex to {xxx}
        if "<" in openapi_path and ">" in openapi_path:
            new_path_list: list = []
            for sub_path in openapi_path.split("/"):
                if not sub_path:
                    continue

                l_index: int = sub_path.find("<")
                r_index: int = sub_path.find(">")
                if l_index != -1 and r_index != -1:
                    real_sub_path = "{" + sub_path[l_index + 1 : r_index] + "}"
                else:
                    real_sub_path = sub_path
                new_path_list.append(real_sub_path)
            openapi_path = "/".join(new_path_list)
        if not openapi_path.startswith("/"):
            openapi_path = "/" + openapi_path
        if openapi_path.endswith("$"):
            openapi_path = openapi_path[:-1]
        base_name: str = rule.target.__name__
        for method in ["get", "post", "head", "options", "delete", "put", "trace", "patch"]:
            handler: Optional[Callable] = getattr(rule.target, method, None)
            if not handler:
                continue
            route_name: str = f"{base_name}.{method}"
            pait_id: Optional[str] = getattr(handler, "_pait_id", None)
            if not pait_id:
                continue
            pait_data.add_route_info(
                AppHelper.app_name, pait_id, path, openapi_path, {method}, route_name, project_name
            )
            _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
    return _pait_data



import logging
from typing import Callable, Dict, Optional, Set

from flask import Flask
from flask.views import MethodView

from pait.g import pait_data
from pait.model.core import PaitCoreModel

from ._app_helper import AppHelper

__all__ = ["load_app"]


def load_app(app: Flask, project_name: str = "") -> Dict[str, PaitCoreModel]:
    """Read data from the route that has been registered to `pait`"""
    _pait_data: Dict[str, PaitCoreModel] = {}
    if not project_name:
        project_name = app.import_name.split(".")[0]
    for route in app.url_map.iter_rules():
        path: str = route.rule
        method_set: Set[str] = route.methods
        route_name: str = route.endpoint
        endpoint: Callable = app.view_functions[route_name]
        pait_id: Optional[str] = getattr(endpoint, "_pait_id", None)
        # replace path <xxx> to {xxx}
        openapi_path: str = path
        if "<" in openapi_path and ">" in openapi_path:
            new_path_list: list = []
            for sub_path in openapi_path.split("/"):
                if not sub_path:
                    continue
                if sub_path[0] == "<" and sub_path[-1] == ">":
                    real_sub_path: str = sub_path[1:-1]
                    if ":" in real_sub_path:
                        real_sub_path = real_sub_path.split(":")[0]
                    real_sub_path = "{" + real_sub_path + "}"
                else:
                    real_sub_path = sub_path
                new_path_list.append(real_sub_path)
            openapi_path = "/".join(new_path_list)
        if not openapi_path.startswith("/"):
            openapi_path = "/" + openapi_path
        if not pait_id:
            if route_name == "static":
                continue
            view_class_endpoint = getattr(endpoint, "view_class", None)
            if not view_class_endpoint or not issubclass(view_class_endpoint, MethodView):
                logging.warning(
                    f"loan path:{path} fail, endpoint:{endpoint} not `view_class` attributes"
                )  # pragma: no cover
                continue
            for method in view_class_endpoint.methods:
                method = method.lower()
                method_set = {method}
                endpoint = getattr(view_class_endpoint, method, None)
                if not endpoint:
                    continue
                pait_id = getattr(endpoint, "_pait_id", None)
                if not pait_id:
                    continue
                pait_data.add_route_info(
                    AppHelper.app_name, pait_id, path, openapi_path, method_set, f"{route_name}.{method}", project_name
                )
                _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
        else:
            pait_data.add_route_info(
                AppHelper.app_name, pait_id, path, openapi_path, method_set, route_name, project_name
            )
            _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
    return _pait_data



from typing import Any, Dict, Type

from pait.model.core import PaitCoreModel
from pait.model.response import PaitBaseResponseModel, PaitJsonResponseModel
from pait.plugin.base import BaseAsyncPlugin, BasePlugin, PluginProtocol
from pait.util import get_pait_response_model


class AutoCompleteJsonRespPluginProtocolMixin(PluginProtocol):
    is_pre_core: bool = False

    def __init__(self, *, pait_response_model: PaitJsonResponseModel) -> None:
        super(AutoCompleteJsonRespPluginProtocolMixin, self).__init__()
        self.pait_response_model: PaitJsonResponseModel = pait_response_model

    def update(self, source_dict: dict, target_dict: dict) -> None:
        for key, value in source_dict.items():
            if isinstance(value, dict) and key in target_dict:
                self.update(value, target_dict[key])
            else:
                source_dict[key] = target_dict.get(key, value)

    @classmethod
    def cls_hook_by_core_model(cls, pait_core_model: PaitCoreModel, kwargs: Dict) -> Dict:
        super().cls_hook_by_core_model(pait_core_model, kwargs)
        if "pait_response_model" in kwargs:
            raise RuntimeError("Please use response_model_list param")
        pait_response_model: Type[PaitBaseResponseModel] = get_pait_response_model(
            pait_core_model.response_model_list, find_core_response_model=True
        )
        if not issubclass(pait_response_model, PaitJsonResponseModel):
            raise ValueError(f"pait_response_model must {PaitJsonResponseModel} not {pait_response_model}")
        kwargs["pait_response_model"] = pait_response_model
        return kwargs


class AutoCompleteJsonRespPlugin(AutoCompleteJsonRespPluginProtocolMixin, BasePlugin):
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        default_response_dict: dict = self.pait_response_model.get_default_dict()
        response_dict: dict = self.call_next(*args, **kwargs)
        self.update(default_response_dict, response_dict)
        return default_response_dict


class AsyncAutoCompleteJsonRespPlugin(AutoCompleteJsonRespPluginProtocolMixin, BaseAsyncPlugin):
    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        default_response_dict: dict = self.pait_response_model.get_default_dict()
        response_dict: dict = await self.call_next(*args, **kwargs)
        self.update(default_response_dict, response_dict)
        return default_response_dict



import json
import logging
from typing import Any, Dict, List, Optional

from sanic import response
from sanic.app import Sanic
from sanic.blueprints import Blueprint
from sanic.exceptions import NotFound
from sanic.request import Request
from sanic.response import HTTPResponse
from sanic_testing.testing import SanicTestClient, TestingResponse  # type: ignore

from pait.api_doc.html import get_redoc_html as _get_redoc_html
from pait.api_doc.html import get_swagger_ui_html as _get_swagger_ui_html
from pait.api_doc.open_api import PaitOpenApi
from pait.field import Depends, Query
from pait.g import config
from pait.model.core import PaitCoreModel
from pait.model.status import PaitStatus
from pait.model.tag import Tag

from ._load_app import load_app
from ._pait import Pait

__all__ = ["add_doc_route"]


def add_doc_route(
    app: Sanic,
    scheme: Optional[str] = None,
    open_json_url_only_path: bool = False,
    prefix: str = "/",
    pin_code: str = "",
    title: str = "Pait Doc",
    open_api_tag_list: Optional[List[Dict[str, Any]]] = None,
) -> None:
    if pin_code:
        logging.info(f"doc route start pin code:{pin_code}")

    doc_pait: Pait = Pait(
        author=config.author or ("so1n",),
        status=config.status or PaitStatus.release,
        tag=(Tag("pait_doc", desc="pait default doc route"),),
        group="pait_doc",
    )

    def _get_request_pin_code(r_pin_code: str = Query.i("", alias="pin_code")) -> Optional[str]:
        if pin_code:
            if r_pin_code != pin_code:
                raise NotFound("")
        return r_pin_code

    def _get_open_json_url(request: Request, r_pin_code: str) -> str:
        _scheme: str = scheme or request.scheme
        if open_json_url_only_path:
            openapi_json_url: str = f"{'/'.join(request.path.split('/')[:-1])}/openapi.json"
        else:
            openapi_json_url = f"{_scheme}://{request.host}{'/'.join(request.path.split('/')[:-1])}/openapi.json"
        if r_pin_code:
            openapi_json_url += f"?pin_code={r_pin_code}"
        return openapi_json_url

    @doc_pait()
    def get_redoc_html(request: Request, r_pin_code: str = Depends.i(_get_request_pin_code)) -> HTTPResponse:
        return response.html(_get_redoc_html(_get_open_json_url(request, r_pin_code), title))

    @doc_pait()
    def get_swagger_ui_html(request: Request, r_pin_code: str = Depends.i(_get_request_pin_code)) -> HTTPResponse:
        return response.html(_get_swagger_ui_html(_get_open_json_url(request, r_pin_code), title))

    @doc_pait(pre_depend_list=[_get_request_pin_code])
    def openapi_route(request: Request) -> HTTPResponse:
        pait_dict: Dict[str, PaitCoreModel] = load_app(request.app)
        _scheme: str = scheme or request.scheme
        pait_openapi: PaitOpenApi = PaitOpenApi(
            pait_dict,
            title=title,
            open_api_server_list=[{"url": f"{_scheme}://{request.host}", "description": ""}],
            open_api_tag_list=open_api_tag_list,
        )
        return response.json(pait_openapi.open_api_dict, dumps=json.dumps, cls=config.json_encoder)

    blueprint: Blueprint = Blueprint("api doc", prefix)
    blueprint.add_route(get_redoc_html, "/redoc", methods={"GET"})
    blueprint.add_route(get_swagger_ui_html, "/swagger", methods={"GET"})
    blueprint.add_route(openapi_route, "/openapi.json", methods={"GET"})
    app.blueprint(blueprint)



from typing import Type

import aiofiles  # type: ignore
from sanic_testing.testing import SanicTestClient, TestingResponse  # type: ignore

from pait.app.base import BaseAppHelper
from pait.core import Pait as _Pait

from ._app_helper import AppHelper

__all__ = ["pait", "Pait"]


class Pait(_Pait):
    app_helper_class: "Type[BaseAppHelper]" = AppHelper


pait = Pait()



from typing import Any

import aiofiles  # type: ignore
from tornado.web import RequestHandler

from pait.model import response
from pait.plugin.base_mock_response import BaseAsyncMockPlugin


class MockPlugin(BaseAsyncMockPlugin):
    tornado_handle: RequestHandler

    def mock_response(self) -> Any:
        async def make_mock_response() -> Any:
            self.tornado_handle.set_status(self.pait_response_model.status_code[0])
            for key, value in self.pait_response_model.header.items():
                self.tornado_handle.set_header(key, value)
            self.tornado_handle.set_header("Content-Type", self.pait_response_model.media_type)
            if issubclass(self.pait_response_model, response.PaitJsonResponseModel):
                self.tornado_handle.write(self.pait_response_model.get_example_value())
            elif issubclass(self.pait_response_model, response.PaitTextResponseModel) or issubclass(
                self.pait_response_model, response.PaitHtmlResponseModel
            ):
                self.tornado_handle.write(self.pait_response_model.get_example_value())
            elif issubclass(self.pait_response_model, response.PaitFileResponseModel):
                async with aiofiles.tempfile.NamedTemporaryFile() as f:  # type: ignore
                    await f.write(self.pait_response_model.get_example_value())
                    await f.seek(0)
                    async for line in f:
                        self.tornado_handle.write(line)
            else:
                raise NotImplementedError()

        return make_mock_response()

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        self.tornado_handle = args[0]
        await super().__call__(args, kwargs)



from typing import TYPE_CHECKING, Any

from pydantic import BaseConfig

if TYPE_CHECKING:
    from pait.model.config import Config
    from pait.model.core import PaitCoreModel


def sync_config_data_to_pait_core_model(config: "Config", pait_core_model: "PaitCoreModel", **kwargs: Any) -> None:
    if not pait_core_model.author:
        pait_core_model.author = config.author
    if not pait_core_model.status:
        pait_core_model.status = config.status

    if config.default_response_model_list:
        pait_core_model.add_response_model_list(config.default_response_model_list)
    if config.default_pydantic_model_config and pait_core_model.pydantic_model_config != BaseConfig:
        pait_core_model.pydantic_model_config = config.default_pydantic_model_config
    if config.block_http_method_set:
        pait_core_model.block_http_method_set = config.block_http_method_set
        pait_core_model.method_list = pait_core_model.method_list



from typing import Any, Optional

from starlette.responses import JSONResponse

from pait.model.response import PaitJsonResponseModel
from pait.plugin.auto_complete_json_resp import AsyncAutoCompleteJsonRespPlugin as _AsyncAutoCompleteJsonRespPlugin
from pait.plugin.auto_complete_json_resp import AutoCompleteJsonRespPlugin as _AutoCompleteJsonRespPlugin

__all__ = ["AutoCompleteJsonRespPlugin", "AsyncAutoCompleteJsonRespPlugin"]


class AsyncAutoCompleteJsonRespPlugin(_AsyncAutoCompleteJsonRespPlugin):
    def __init__(
        self,
        *,
        pait_response_model: PaitJsonResponseModel,
        status_code: int = 200,
        headers: Optional[dict] = None,
        media_type: Optional[str] = None,
    ) -> None:
        super(AsyncAutoCompleteJsonRespPlugin, self).__init__(pait_response_model=pait_response_model)
        self.status_code: int = status_code
        self.headers: Optional[dict] = headers
        self.media_type: Optional[str] = media_type

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        response_dict: Any = await super(AsyncAutoCompleteJsonRespPlugin, self).__call__(*args, **kwargs)
        return JSONResponse(
            response_dict, status_code=self.status_code, headers=self.headers, media_type=self.media_type
        )


class AutoCompleteJsonRespPlugin(_AutoCompleteJsonRespPlugin):
    def __init__(
        self,
        *,
        pait_response_model: PaitJsonResponseModel,
        status_code: int = 200,
        headers: Optional[dict] = None,
        media_type: Optional[str] = None,
    ) -> None:
        super(AutoCompleteJsonRespPlugin, self).__init__(pait_response_model=pait_response_model)
        self.status_code: int = status_code
        self.headers: Optional[dict] = headers
        self.media_type: Optional[str] = media_type

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        response_dict: Any = super(AutoCompleteJsonRespPlugin, self).__call__(*args, **kwargs)
        return JSONResponse(
            response_dict, status_code=self.status_code, headers=self.headers, media_type=self.media_type
        )



from collections.abc import AsyncIterator, Iterator
from typing import Any, List, Optional, Set, Type, Union, _GenericAlias  # type: ignore

from pait.exceptions import ParseTypeError

_CAN_JSON_TYPE_SET: Set[Optional[type]] = {bool, dict, float, int, list, str, tuple, type(None), None}


def parse_typing(_type: Any) -> Union[List[Type[Any]], Type]:
    """
    parse typing.type to Python.type
    >>> from typing import Dict, Optional
    >>> assert dict is parse_typing(dict)
    >>> assert list is parse_typing(List)
    >>> assert dict is parse_typing(Dict)
    >>> assert dict in set(parse_typing(Optional[Dict]))
    >>> assert None in set(parse_typing(Optional[Dict]))
    >>> assert dict in set(parse_typing(Optional[dict]))
    >>> assert None in set(parse_typing(Optional[dict]))
    >>> assert dict is parse_typing(Union[dict])
    >>> assert dict is parse_typing(Union[Dict])
    >>> assert dict is parse_typing(Union[Dict[str, Any]])
    """
    if isinstance(_type, _GenericAlias):
        # support typing.xxx
        origin: type = _type.__origin__  # get typing.xxx's raw type
        if origin is Union:
            # support Union, Optional
            type_list: List[Type[Any]] = []
            for i in _type.__args__:
                if isinstance(i, list):
                    for j in i:
                        value: Union[List[Type[Any]], Type] = parse_typing(j)
                        if isinstance(value, list):
                            type_list.extend(value)
                        else:
                            type_list.append(value)
                else:
                    value = parse_typing(i)
                    if isinstance(value, list):
                        type_list.extend(value)
                    else:
                        type_list.append(value)
            return type_list
        elif origin in (AsyncIterator, Iterator):
            # support AsyncIterator, Iterator
            return parse_typing(_type.__args__[0])
        return origin
    elif _type in _CAN_JSON_TYPE_SET:
        return _type
    else:
        raise ParseTypeError(f"Can not parse {_type} origin type")


def is_type(source_type: Type, target_type: Union[Type, object]) -> bool:
    """Determine whether the two types are consistent"""
    parse_source_type: Union[List[Type], Type] = parse_typing(source_type)
    if not isinstance(parse_source_type, list):
        parse_source_type = [parse_source_type]

    parse_target_type: Union[List[Type], Type] = parse_typing(target_type)
    if not isinstance(parse_target_type, list):
        parse_target_type = [parse_target_type]
    return bool(set(parse_target_type) & set(parse_source_type))



import copy
import inspect
import logging
from typing import TYPE_CHECKING, Callable, List, Optional, Set, Tuple, Type

from pydantic import BaseConfig

from pait.model.response import PaitBaseResponseModel, PaitResponseModel
from pait.model.status import PaitStatus
from pait.param_handle import AsyncParamHandler, ParamHandler
from pait.plugin import PluginManager

if TYPE_CHECKING:
    from pait.app.base import BaseAppHelper


class PaitCoreModel(object):
    def __init__(
        self,
        func: Callable,
        app_helper_class: "Type[BaseAppHelper]",
        pre_depend_list: Optional[List[Callable]] = None,
        path: Optional[str] = None,
        openapi_path: Optional[str] = None,
        method_set: Optional[Set[str]] = None,
        operation_id: Optional[str] = None,
        func_name: Optional[str] = None,
        author: Optional[Tuple[str, ...]] = None,
        summary: Optional[str] = None,
        desc: Optional[str] = None,
        status: Optional[PaitStatus] = None,
        group: Optional[str] = None,
        tag: Optional[Tuple[str, ...]] = None,
        response_model_list: Optional[List[Type[PaitBaseResponseModel]]] = None,
        pydantic_model_config: Optional[Type[BaseConfig]] = None,
        plugin_list: Optional[List[PluginManager]] = None,
        post_plugin_list: Optional[List[PluginManager]] = None,
    ):
        # pait
        self.app_helper_class: "Type[BaseAppHelper]" = app_helper_class
        self.func: Callable = func  # route func
        self.qualname: str = func.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0]
        self.pait_id: str = f"{self.qualname}_{id(func)}"
        setattr(func, "_pait_id", self.pait_id)
        setattr(func, "pait_core_model", self)
        self.pre_depend_list: List[Callable] = pre_depend_list or []
        self.func_path: str = ""
        self.block_http_method_set: Set[str] = set()
        self.pydantic_model_config: Type[BaseConfig] = pydantic_model_config or BaseConfig

        # api doc
        self.path: str = path or ""  # request url path
        self.openapi_path: str = openapi_path or ""
        self._method_list: List[str] = sorted(list(method_set or set()))  # request method set
        self.func_name: str = func_name or func.__name__
        self.operation_id: str = operation_id or self.func_name  # route name
        self.author: Optional[Tuple[str, ...]] = author  # The main developer of this func
        self.summary: str = summary or ""
        self.desc: str = desc or func.__doc__ or ""  # desc of this func
        self.status: Optional[PaitStatus] = status or PaitStatus.undefined
        self.group: str = group or "root"  # Which group this interface belongs to
        self.tag: Tuple[str, ...] = tag or ("default",)  # Interface tag

        self._response_model_list: List[Type[PaitBaseResponseModel]] = []
        if response_model_list:
            self.add_response_model_list(response_model_list)

        # pait plugin
        self._plugin_list: List[PluginManager] = []
        self._post_plugin_list: List[PluginManager] = []
        self._plugin_manager_list: List[PluginManager] = []
        if inspect.iscoroutinefunction(self.func):
            self._param_handler_plugin: PluginManager = PluginManager(AsyncParamHandler)
        else:
            self._param_handler_plugin = PluginManager(ParamHandler)
        self._param_handler_plugin.cls_hook_by_core_model(self)
        self.add_plugin(plugin_list, post_plugin_list)

    @property
    def method_list(self) -> List[str]:
        return self._method_list

    @method_list.setter
    def method_list(self, method_list: List[str]) -> None:
        _temp_set: Set[str] = set(self._method_list) | set(method_list)
        _temp_set.difference_update(self.block_http_method_set)
        self._method_list = sorted(list(_temp_set))

    @property
    def response_model_list(self) -> List[Type[PaitBaseResponseModel]]:
        return self._response_model_list

    def add_response_model_list(self, response_model_list: List[Type[PaitBaseResponseModel]]) -> None:
        for response_model in response_model_list:
            if response_model in self._response_model_list:
                continue
            if issubclass(response_model, PaitResponseModel):
                logging.warning(
                    f"Please replace {self.operation_id}'s response model {response_model}"
                    f" with {PaitBaseResponseModel}"
                )
            self._response_model_list.append(response_model)

    @property
    def plugin_list(self) -> List[PluginManager]:
        return self._plugin_manager_list

    def add_plugin(
        self, plugin_list: Optional[List[PluginManager]], post_plugin_list: Optional[List[PluginManager]]
    ) -> None:
        raw_plugin_list: List[PluginManager] = self._plugin_list
        raw_post_plugin_list: List[PluginManager] = self._post_plugin_list
        try:
            for plugin_manager in plugin_list or []:
                plugin_manager.cls_hook_by_core_model(self)
                if not plugin_manager.plugin_class.is_pre_core:
                    raise ValueError(f"{plugin_manager.plugin_class} is post plugin")
                self._plugin_list.append(plugin_manager)

            for plugin_manager in post_plugin_list or []:
                plugin_manager.cls_hook_by_core_model(self)
                if plugin_manager.plugin_class.is_pre_core:
                    raise ValueError(f"{plugin_manager.plugin_class} is pre plugin")
                self._post_plugin_list.append(plugin_manager)
        except Exception as e:
            self._plugin_list = raw_plugin_list
            self._post_plugin_list = raw_post_plugin_list
            raise e
        else:
            # In future version, it may be possible to switch plugins at runtime
            plugin_manager_list: List[PluginManager] = copy.deepcopy(
                self._plugin_list + [self._param_handler_plugin] + self._post_plugin_list
            )
            plugin_manager_list.reverse()
            self._plugin_manager_list = plugin_manager_list



from typing import Dict

__all__ = ["Tag"]
_tag_dict: Dict[str, str] = {}


class Tag(object):
    def __init__(self, name: str, desc: str = ""):
        self._name: str = name
        self._desc: str = desc
        if self._name in _tag_dict and _tag_dict[self._name] != self._desc:
            raise KeyError(f"{self._name} tag already exists")

        _tag_dict[self._name] = self.desc

    @property
    def name(self) -> str:
        return self._name

    @property
    def desc(self) -> str:
        return self._desc



import inspect
from typing import TYPE_CHECKING, Any, Dict, Type, TypeVar, Union

if TYPE_CHECKING:
    from pait.model.core import PaitCoreModel

_PluginT = TypeVar("_PluginT", bound="PluginProtocol")


class PluginProtocol(object):
    is_pre_core: bool = True
    pait_core_model: "PaitCoreModel"
    args: list
    kwargs: dict

    def __init__(self, **kwargs: Any) -> None:
        pass

    @classmethod
    def cls_hook_by_core_model(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        return kwargs  # pragma: no cover

    def __post_init__(self, pait_core_model: "PaitCoreModel", args: tuple, kwargs: dict) -> None:
        self.pait_core_model = pait_core_model
        self.args = list(args) or []
        self.kwargs = kwargs or {}


class BasePlugin(PluginProtocol):
    def call_next(self, *args: Any, **kwargs: Any) -> Any:
        raise RuntimeError("Failed to load PluginManager, please check the list of plugins")  # pragma: no cover

    @classmethod
    def cls_hook_by_core_model(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        if inspect.iscoroutinefunction(pait_core_model.func):
            raise TypeError("PluginManager not support async func")
        return kwargs

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.call_next(*args, **kwargs)  # pragma: no cover


class BaseAsyncPlugin(PluginProtocol):
    async def call_next(self, *args: Any, **kwargs: Any) -> Any:
        raise RuntimeError("Failed to load PluginManager, please check the list of plugins")  # pragma: no cover

    @classmethod
    def cls_hook_by_core_model(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        if not inspect.iscoroutinefunction(pait_core_model.func):
            raise TypeError("PluginManager only support async func")
        return kwargs

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return await self.call_next(*args, **kwargs)  # pragma: no cover


_T = Union[BasePlugin, BaseAsyncPlugin]


class PluginManager(object):
    def __init__(self, plugin_class: Type[_T], **kwargs: Any):
        self.plugin_class: Type[_T] = plugin_class
        self._kwargs: Any = kwargs

    def cls_hook_by_core_model(self, pait_core_model: "PaitCoreModel") -> None:
        self._kwargs = self.plugin_class.cls_hook_by_core_model(pait_core_model, self._kwargs)

    def get_plugin(self) -> _T:
        return self.plugin_class(**self._kwargs)



from typing import Any, Callable, Dict, Optional

from sanic.response import json

from pait.model.response import PaitJsonResponseModel
from pait.plugin.auto_complete_json_resp import AsyncAutoCompleteJsonRespPlugin as _AsyncAutoCompleteJsonRespPlugin

__all__ = ["AsyncAutoCompleteJsonRespPlugin"]


class AsyncAutoCompleteJsonRespPlugin(_AsyncAutoCompleteJsonRespPlugin):
    def __init__(
        self,
        *,
        pait_response_model: PaitJsonResponseModel,
        status: int = 200,
        headers: Optional[Dict[str, str]] = None,
        content_type: str = "application/json",
        dumps: Optional[Callable[..., str]] = None,
        **json_kwargs: Any,
    ) -> None:
        super(AsyncAutoCompleteJsonRespPlugin, self).__init__(pait_response_model=pait_response_model)
        self.status: int = status
        self.headers: Optional[Dict[str, str]] = headers
        self.content_type = content_type
        self.dumps: Optional[Callable[..., str]] = dumps
        self.json_kwargs: dict = json_kwargs

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        response_dict: Any = await super(AsyncAutoCompleteJsonRespPlugin, self).__call__(*args, **kwargs)
        return json(
            response_dict,
            status=self.status,
            headers=self.headers,
            content_type=self.content_type,
            dumps=self.dumps,
            **self.json_kwargs,
        )



import copy
import difflib
from typing import Any, Callable, Dict, Generic, List, Mapping, Optional, Type, TypeVar
from urllib.parse import urlencode

from pydantic import BaseModel, ValidationError

from pait.model import response
from pait.model.core import PaitCoreModel
from pait.util import gen_example_dict_from_schema, gen_example_value_from_python, get_pait_response_model

RESP_T = TypeVar("RESP_T")


class BaseTestHelper(Generic[RESP_T]):
    client: Any

    def __init__(
        self,
        client: Any,
        func: Callable,
        pait_dict: Optional[Dict[str, PaitCoreModel]] = None,
        body_dict: Optional[dict] = None,
        cookie_dict: Optional[dict] = None,
        file_dict: Optional[dict] = None,
        form_dict: Optional[dict] = None,
        header_dict: Optional[dict] = None,
        path_dict: Optional[dict] = None,
        query_dict: Optional[dict] = None,
        strict_inspection_check_json_content: bool = True,
        find_coro_response_model: bool = False,
        target_pait_response_class: Optional[Type["response.PaitBaseResponseModel"]] = None,
    ):
        """
        :param client:  test client
        :param func:  route handle decorated by pait
        :param pait_dict:  pait core data
        :param body_dict:  request body param
        :param cookie_dict:  request cookie param
        :param file_dict:  request file param
        :param form_dict:  request form param
        :param header_dict:  request herder param
        :param path_dict:  request path param
        :param query_dict:  request query param
        """
        pait_id: str = getattr(func, "_pait_id", "")
        if not pait_id:
            raise RuntimeError(f"Can not found pait id from {func}")

        self.client: Any = client
        self.func: Callable = func
        # pait dict handle
        if pait_dict:
            self.pait_dict: Dict[str, PaitCoreModel] = pait_dict
        else:
            self.pait_dict = self._gen_pait_dict()
        self.pait_core_model: PaitCoreModel = self.pait_dict[pait_id]

        self.body_dict: Optional[dict] = body_dict
        self.cookie_dict: Optional[dict] = cookie_dict
        self.file_dict: Optional[dict] = file_dict
        self.form_dict: Optional[dict] = form_dict
        self.header_dict: dict = header_dict or {}
        self.path_dict: Optional[dict] = path_dict
        self.query_dict: Optional[dict] = query_dict
        self.strict_inspection_check_json_content: bool = strict_inspection_check_json_content

        # path handle
        self.path: str = self.pait_core_model.path
        if self.path_dict:
            path_list: List[str] = self.path.split("/")
            new_path_list: List[str] = []
            for sub_path in path_list:
                if not sub_path:
                    continue
                new_sub_path: Optional[str] = self._replace_path(sub_path)
                if new_sub_path:
                    new_path_list.append(new_sub_path)
                else:
                    new_path_list.append(sub_path)
            self.path = "/".join([str(i) for i in new_path_list])

        # add query path by path
        if self.query_dict:
            self.path = self.path + "?" + urlencode(self.query_dict, True)
        if not self.path.startswith("/"):
            self.path = "/" + self.path

        self.find_coro_response_model: bool = find_coro_response_model
        self.target_pait_response_class: Optional[Type["response.PaitBaseResponseModel"]] = target_pait_response_class
        self._app_init_field()

    def _app_init_field(self) -> None:
        """init request param by application framework"""
        raise NotImplementedError()

    def _gen_pait_dict(self) -> Dict[str, PaitCoreModel]:
        """load pait dict"""
        raise NotImplementedError()

    @staticmethod
    def _get_status_code(resp: RESP_T) -> int:
        """get response status code"""
        raise NotImplementedError()

    @staticmethod
    def _get_content_type(resp: RESP_T) -> str:
        """get response content type"""
        raise NotImplementedError()

    @staticmethod
    def _get_json(resp: RESP_T) -> dict:
        """get json response"""
        raise NotImplementedError()

    @staticmethod
    def _get_headers(resp: RESP_T) -> Mapping:
        """get response header"""
        raise NotImplementedError()

    @staticmethod
    def _get_text(resp: RESP_T) -> str:
        """get text response"""
        raise NotImplementedError()

    @staticmethod
    def _get_bytes(resp: RESP_T) -> bytes:
        """get bytes response"""
        raise NotImplementedError()

    def _check_diff_resp_dict(
        self, raw_resp: Any, default_resp: Any, parent_key_list: Optional[List[str]] = None
    ) -> bool:
        """check resp data structure diff, Only return True to pass the verification"""
        raw_parent_key_list: List[str] = parent_key_list or []
        try:
            if isinstance(raw_resp, dict):
                for key, value in raw_resp.items():
                    parent_key_list = copy.deepcopy(raw_parent_key_list)
                    parent_key_list.append(key)
                    if not self._check_diff_resp_dict(value, default_resp[key], parent_key_list):
                        return False
                return True
            elif isinstance(raw_resp, list) or isinstance(raw_resp, tuple):
                if (
                    raw_resp
                    and default_resp
                    and not self._check_diff_resp_dict(raw_resp[0], default_resp[0], parent_key_list)
                ):
                    return False
                return True
            else:
                return True
        except KeyError:
            raise RuntimeError(f"Can not found key from model, key:{' -> '.join(parent_key_list or '')}")

    def _assert_response(self, resp: RESP_T) -> None:
        """Whether the structure of the check response is correct"""
        if not self.pait_core_model.response_model_list:
            return

        real_response_model: Optional[Type[response.PaitBaseResponseModel]] = None
        max_quick_ratio: float = 0.0
        model_check_msg_dict: Dict[Type[response.PaitBaseResponseModel], List[str]] = {}
        response_model_list: List[Type[response.PaitBaseResponseModel]] = (
            self.pait_core_model.response_model_list
            if not self.find_coro_response_model
            else [
                get_pait_response_model(
                    self.pait_core_model.response_model_list,
                    target_pait_response_class=self.target_pait_response_class,
                    find_core_response_model=self.find_coro_response_model,
                )
            ]
        )
        for response_model in response_model_list:
            error_msg_list: List[str] = []
            if self._get_status_code(resp) not in response_model.status_code:
                error_msg_list.append("check status code error.")
            if response_model.media_type not in self._get_content_type(resp):
                error_msg_list.append("check media type error.")

            resp_header: Mapping = self._get_headers(resp)
            for key, _ in response_model.header.items():
                if key not in resp_header:
                    error_msg_list.append(f"check header error. Can not found header:{key} in {resp_header}")

            # check content
            if issubclass(response_model, response.PaitJsonResponseModel):
                response_data_model: Type[BaseModel] = response_model.response_data
                resp_dict: Optional[dict] = self._get_json(resp)
                if not resp_dict:
                    continue
                response_data_default_dict: dict = gen_example_dict_from_schema(response_data_model.schema())
                ratio: float = difflib.SequenceMatcher(
                    None,
                    str(gen_example_value_from_python(resp_dict)),
                    str(response_data_default_dict),
                ).quick_ratio()
                if ratio > max_quick_ratio:
                    max_quick_ratio = ratio
                    real_response_model = response_model
                try:
                    response_data_model(**resp_dict)
                    if self.strict_inspection_check_json_content and not self._check_diff_resp_dict(
                        resp_dict, response_data_default_dict
                    ):
                        error_msg_list.append("check json structure error")
                except (ValidationError, RuntimeError) as e:
                    error_msg_list.append(f"check json content error, exec: {e}")

            elif issubclass(response_model, response.PaitHtmlResponseModel) or issubclass(
                response_model, response.PaitTextResponseModel
            ):
                real_response_model = response_model
                if not isinstance(self._get_text(resp), type(response_model.response_data)):
                    error_msg_list.append("check text content type error")  # pragma: no cover
            elif issubclass(response_model, response.PaitFileResponseModel):
                real_response_model = response_model
                if not isinstance(self._get_bytes(resp), type(response_model.response_data)):
                    error_msg_list.append("check bytes content type error")  # pragma: no cover
            else:
                raise TypeError(f"Pait not support response model:{response_model}")
            if not error_msg_list:
                return
            else:
                model_check_msg_dict[response_model] = error_msg_list
        response_dict: dict = {
            "status_code": self._get_status_code(resp),
            "media_type": self._get_content_type(resp),
            "headers": self._get_headers(resp),
            "obj": resp,
        }
        if real_response_model in model_check_msg_dict:
            raise RuntimeError(
                f"Check Response Error."
                f" maybe error result: {model_check_msg_dict[real_response_model]}"
                f" by response model:{real_response_model}"
                f" response info: {response_dict}"
            )
        raise RuntimeError(
            "Check Response Error. "
            f"response error result:{model_check_msg_dict} "
            f"total response model:{self.pait_core_model.response_model_list}. "
            f"response info: {response_dict} "
        )  # pragma: no cover

    def _replace_path(self, path_str: str) -> Optional[str]:
        raise NotImplementedError()

    def _real_request(self, method: str) -> RESP_T:
        """Whether the structure of the check response is correct"""
        raise NotImplementedError()

    def request(self, method: Optional[str] = None) -> RESP_T:
        """user call test request api"""
        if not method:
            # auto select method
            if len(self.pait_core_model.method_list) == 1:
                method = self.pait_core_model.method_list[0]
            else:
                raise RuntimeError(
                    f"Pait Can not auto select method, please choice method in {self.pait_core_model.method_list}"
                )
        resp: RESP_T = self._real_request(method)
        if self.pait_core_model.response_model_list:
            self._assert_response(resp)
        return resp

    def json(self, method: Optional[str] = None) -> dict:
        return self._get_json(self.request(method))

    def text(self, method: Optional[str] = None) -> str:
        return self._get_text(self.request(method))

    def get(self) -> RESP_T:
        return self.request("GET")

    def patch(self) -> RESP_T:
        return self.request("PATCH")

    def post(self) -> RESP_T:
        return self.request("POST")

    def head(self) -> RESP_T:
        return self.request("HEAD")

    def put(self) -> RESP_T:
        return self.request("PUT")

    def delete(self) -> RESP_T:
        return self.request("DELETE")

    def options(self) -> RESP_T:
        return self.request("OPTIONS")



from dataclasses import MISSING
from typing import TYPE_CHECKING, Any, Callable, Optional

from pydantic.fields import FieldInfo, Undefined
from pydantic.typing import NoArgAnyCallable

if TYPE_CHECKING:
    from pait.model.links import LinksModel


class BaseField(FieldInfo):
    field_name: str = ""
    media_type: str = "*/*"
    openapi_serialization: Optional[dict] = None

    def __init__(
        self,
        default: Any = Undefined,
        *,
        link: "Optional[LinksModel]" = None,
        media_type: str = "",
        openapi_serialization: Any = None,
        raw_return: bool = False,
        example: Any = MISSING,
        default_factory: Optional[NoArgAnyCallable] = None,
        alias: str = None,
        title: str = None,
        description: str = None,
        const: bool = None,
        gt: float = None,
        ge: float = None,
        lt: float = None,
        le: float = None,
        multiple_of: float = None,
        min_items: int = None,
        max_items: int = None,
        min_length: int = None,
        max_length: int = None,
        regex: str = None,
        **extra: Any,
    ):
        if self.__class__.__mro__[2] != FieldInfo:
            raise RuntimeError("Only classes that inherit BaseField can be used")
        self.raw_return = raw_return
        if example is not MISSING:
            extra["example"] = example

        if not link:
            link = extra.pop("link", None)
        self.link: "Optional[LinksModel]" = link

        self.media_type = media_type or self.__class__.media_type
        self.openapi_serialization = openapi_serialization or self.__class__.openapi_serialization
        super().__init__(
            default,
            default_factory=default_factory,
            alias=alias,
            title=title,
            description=description,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            min_items=min_items,
            max_items=max_items,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            **extra,
        )

    @classmethod
    def get_field_name(cls) -> str:
        return cls.field_name or cls.__name__.lower()

    @classmethod
    def i(
        cls,
        default: Any = Undefined,
        *,
        link: "Optional[LinksModel]" = None,
        raw_return: bool = False,
        media_type: str = "",
        example: Any = MISSING,
        openapi_serialization: Any = None,
        default_factory: Optional[NoArgAnyCallable] = None,
        alias: str = None,
        title: str = None,
        description: str = None,
        const: bool = None,
        gt: float = None,
        ge: float = None,
        lt: float = None,
        le: float = None,
        multiple_of: float = None,
        min_items: int = None,
        max_items: int = None,
        min_length: int = None,
        max_length: int = None,
        regex: str = None,
        **extra: Any,
    ) -> Any:
        """ignore mypy tip"""
        return cls(
            default,
            raw_return=raw_return,
            link=link,
            example=example,
            media_type=media_type,
            openapi_serialization=openapi_serialization,
            default_factory=default_factory,
            alias=alias,
            title=title,
            description=description,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            min_items=min_items,
            max_items=max_items,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            **extra,
        )

    @classmethod
    def from_pydantic_field(cls, field: FieldInfo) -> "BaseField":
        """use replace pydantic property field"""
        return cls(
            field.default,
            default_factory=field.default_factory,
            alias=field.alias,
            title=field.title,
            description=field.description,
            const=field.const,
            gt=field.gt,
            ge=field.ge,
            lt=field.lt,
            le=field.le,
            multiple_of=field.multiple_of,
            min_items=field.min_items,
            max_items=field.max_items,
            min_length=field.min_length,
            max_length=field.max_length,
            regex=field.regex,
            **field.extra,
        )


class Body(BaseField):
    media_type: str = "application/json"


class Cookie(BaseField):
    pass


class File(BaseField):
    media_type: str = "multipart/form-data"


class Form(BaseField):
    media_type: str = "application/x-www-form-urlencoded"


class MultiForm(BaseField):
    media_type: str = "application/x-www-form-urlencoded"
    openapi_serialization: Optional[dict] = {"style": "form", "explode": True}


class Header(BaseField):
    pass


class Path(BaseField):
    pass


class Query(BaseField):
    pass


class MultiQuery(BaseField):
    pass


class Depends(object):
    def __init__(self, func: Callable):
        self.func: Callable = func

    @classmethod
    def i(cls, func: Callable) -> Any:
        return cls(func)


def is_pait_field(pait_field: Any) -> bool:
    return isinstance(pait_field, (BaseField, Depends))



from pait.plugin.at_most_one_of import AsyncAtMostOneOfPlugin, AtMostOneOfPluginProtocol
from pait.plugin.required import AsyncRequiredPlugin, RequiredPlugin

__all__ = ["RequiredPlugin", "AsyncRequiredPlugin", "AsyncAtMostOneOfPlugin", "AtMostOneOfPluginProtocol"]



from abc import ABC
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Type

from pait.model.response import PaitBaseResponseModel
from pait.plugin.base import BaseAsyncPlugin, BasePlugin, PluginProtocol
from pait.util import get_pait_response_model

if TYPE_CHECKING:
    from pait.model.core import PaitCoreModel


class MockPluginInitProtocolMixin(PluginProtocol):
    is_pre_core: bool = False

    def __init__(self, pait_response_model: Type[PaitBaseResponseModel], **kwargs: Any):
        self.pait_response_model: Type[PaitBaseResponseModel] = pait_response_model
        super().__init__(**kwargs)

    @classmethod
    def cls_hook_by_core_model(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        super().cls_hook_by_core_model(pait_core_model, kwargs)
        if not pait_core_model.response_model_list:
            raise RuntimeError(f"{pait_core_model.func} can not found response model")
        if "pait_response_model" in kwargs:
            raise RuntimeError("Please use response_model_list param")
        pait_response: Optional[Type[PaitBaseResponseModel]] = None
        enable_mock_response_filter_fn: Optional[Callable] = kwargs.pop("enable_mock_response_filter_fn", None)
        if enable_mock_response_filter_fn and pait_core_model.response_model_list:
            for _pait_response in pait_core_model.response_model_list:
                if enable_mock_response_filter_fn(_pait_response):
                    pait_response = _pait_response
                    break

        if not pait_response:
            pait_response = get_pait_response_model(
                pait_core_model.response_model_list,
                target_pait_response_class=kwargs.pop("target_pait_response_class", False),
                find_core_response_model=kwargs.pop("find_coro_response_model", None),
            )
        kwargs["pait_response_model"] = pait_response
        return kwargs

    def mock_response(self) -> Any:
        raise NotImplementedError()

    def get_mock_response(self) -> Any:
        return self.mock_response()


class BaseMockPlugin(MockPluginInitProtocolMixin, BasePlugin, ABC):
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.get_mock_response()


class BaseAsyncMockPlugin(MockPluginInitProtocolMixin, BaseAsyncPlugin, ABC):
    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return await self.get_mock_response()



from typing import Any, AsyncContextManager

import aiofiles  # type: ignore
from sanic import response as sanic_response
from sanic.response import json as resp_json
from sanic_testing.testing import SanicTestClient, TestingResponse  # type: ignore

from pait.model import response
from pait.plugin.base_mock_response import BaseAsyncMockPlugin


class MockPlugin(BaseAsyncMockPlugin):
    def mock_response(self) -> Any:
        async def make_mock_response() -> sanic_response.BaseHTTPResponse:
            if issubclass(self.pait_response_model, response.PaitJsonResponseModel):
                resp: sanic_response.BaseHTTPResponse = resp_json(self.pait_response_model.get_example_value())
            elif issubclass(self.pait_response_model, response.PaitTextResponseModel) or issubclass(
                self.pait_response_model, response.PaitHtmlResponseModel
            ):
                resp = sanic_response.text(
                    self.pait_response_model.get_example_value(), content_type=self.pait_response_model.media_type
                )
            elif issubclass(self.pait_response_model, response.PaitFileResponseModel):
                named_temporary_file: AsyncContextManager = aiofiles.tempfile.NamedTemporaryFile()  # type: ignore
                f: Any = await named_temporary_file.__aenter__()
                await f.write(self.pait_response_model.get_example_value())
                await f.seek(0)
                resp = await sanic_response.file_stream(f.name, mime_type=self.pait_response_model.media_type)

                raw_streaming_fn = resp.streaming_fn

                async def _streaming_fn(_response: sanic_response.BaseHTTPResponse) -> None:
                    await raw_streaming_fn(_response)
                    await named_temporary_file.__aexit__(None, None, None)

                resp.streaming_fn = _streaming_fn
            else:
                raise NotImplementedError(f"make_mock_response not support {self.pait_response_model}")
            resp.status = self.pait_response_model.status_code[0]
            if self.pait_response_model.header:
                resp.headers.update(self.pait_response_model.header)
            return resp

        return make_mock_response()



from typing import Any, Callable, Dict, Optional, Type

import pydantic

from pait.model.core import PaitCoreModel
from pait.model.response import PaitBaseResponseModel, PaitJsonResponseModel
from pait.plugin.base import BaseAsyncPlugin, BasePlugin, PluginProtocol
from pait.util import gen_example_dict_from_pydantic_base_model, get_pait_response_model, get_real_annotation


class JsonRespPluginProtocolMixin(PluginProtocol):
    def __init__(self, *, check_resp_fn: Callable) -> None:
        super(JsonRespPluginProtocolMixin, self).__init__()
        self.check_resp_fn: Callable = check_resp_fn

    @classmethod
    def cls_hook_by_core_model(cls, pait_core_model: PaitCoreModel, kwargs: Dict) -> Dict:
        kwargs = super().cls_hook_by_core_model(pait_core_model, kwargs)
        if "check_resp_fn" in kwargs:
            raise RuntimeError("Please use response_model_list param")

        pait_response_model: Type[PaitBaseResponseModel] = get_pait_response_model(
            pait_core_model.response_model_list, find_core_response_model=True
        )
        if not issubclass(pait_response_model, PaitJsonResponseModel):
            raise ValueError(f"pait_response_model must {PaitJsonResponseModel} not {pait_response_model}")

        return_type: Optional[Type] = pait_core_model.func.__annotations__.get("return", None)  # type: ignore
        if not return_type:
            raise ValueError(f"Can not found return type by func:{pait_core_model.func}")
        return_type = get_real_annotation(return_type, pait_core_model.func)

        if hasattr(return_type, "__required_keys__"):
            base_model_class: Type[pydantic.BaseModel] = pydantic.create_model_from_typeddict(
                return_type  # type: ignore
            )
            base_model_class(**gen_example_dict_from_pydantic_base_model(pait_response_model.response_data))

        elif not issubclass(return_type, dict):
            raise ValueError(f"Can not found {cls.__name__} support return type")

        def check_resp_by_dict(response_dict: dict) -> None:
            pait_response_model.response_data(**response_dict)  # type: ignore

        kwargs["check_resp_fn"] = check_resp_by_dict
        return kwargs


class CheckJsonRespPlugin(JsonRespPluginProtocolMixin, BasePlugin):
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        response: Any = self.call_next(*args, **kwargs)
        self.check_resp_fn(response)
        return response


class AsyncCheckJsonRespPlugin(JsonRespPluginProtocolMixin, BaseAsyncPlugin):
    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        response: Any = await self.call_next(*args, **kwargs)
        self.check_resp_fn(response)
        return response



from typing import Any

from tornado.web import RequestHandler

from pait.model.response import PaitJsonResponseModel
from pait.plugin.auto_complete_json_resp import AsyncAutoCompleteJsonRespPlugin as _AsyncAutoCompleteJsonRespPlugin

__all__ = ["AsyncAutoCompleteJsonRespPlugin"]


class AsyncAutoCompleteJsonRespPlugin(_AsyncAutoCompleteJsonRespPlugin):
    def __init__(
        self,
        *,
        pait_response_model: PaitJsonResponseModel,
    ) -> None:
        super(AsyncAutoCompleteJsonRespPlugin, self).__init__(pait_response_model=pait_response_model)

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        response_dict: Any = await super(AsyncAutoCompleteJsonRespPlugin, self).__call__(*args, **kwargs)
        tornado_handle: RequestHandler = args[0]
        tornado_handle.write(response_dict)
        return response_dict



import logging
from typing import Any, Dict, List, Optional

from starlette.applications import Starlette
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse
from starlette.routing import Mount, Route

from pait.api_doc.html import get_redoc_html as _get_redoc_html
from pait.api_doc.html import get_swagger_ui_html as _get_swagger_ui_html
from pait.api_doc.open_api import PaitOpenApi
from pait.field import Depends, Query
from pait.g import config
from pait.model.core import PaitCoreModel
from pait.model.status import PaitStatus
from pait.model.tag import Tag

from ._load_app import load_app
from ._pait import Pait


def add_doc_route(
    app: Starlette,
    scheme: Optional[str] = None,
    open_json_url_only_path: bool = False,
    prefix: str = "/",
    pin_code: str = "",
    title: str = "Pait Doc",
    open_api_tag_list: Optional[List[Dict[str, Any]]] = None,
) -> None:
    if pin_code:
        logging.info(f"doc route start pin code:{pin_code}")

    doc_pait: Pait = Pait(
        author=config.author or ("so1n",),
        status=config.status or PaitStatus.release,
        tag=(Tag("pait_doc", desc="pait default doc route"),),
        group="pait_doc",
    )

    def _get_request_pin_code(r_pin_code: str = Query.i("", alias="pin_code")) -> Optional[str]:
        if pin_code:
            if r_pin_code != pin_code:
                raise HTTPException(
                    status_code=404,
                    detail=(
                        "The requested URL was not found on the server. If you entered"
                        " the URL manually please check your spelling and try again."
                    ),
                )
        return r_pin_code

    def _get_open_json_url(request: Request, r_pin_code: str) -> str:
        _scheme: str = scheme or request.url.scheme
        if open_json_url_only_path:
            openapi_json_url: str = f"{'/'.join(request.url.path.split('/')[:-1])}/openapi.json"
        else:
            url: str = f"{_scheme}://{request.url.hostname}"
            if request.url.port:
                url += f":{request.url.port}"  # pragma: no cover
            openapi_json_url = f"{url}{'/'.join(request.url.path.split('/')[:-1])}/openapi.json"
        if r_pin_code:
            openapi_json_url += f"?pin_code={r_pin_code}"
        return openapi_json_url

    @doc_pait()
    def get_redoc_html(request: Request, r_pin_code: str = Depends.i(_get_request_pin_code)) -> HTMLResponse:
        return HTMLResponse(_get_redoc_html(_get_open_json_url(request, r_pin_code), title))

    @doc_pait()
    def get_swagger_ui_html(request: Request, r_pin_code: str = Depends.i(_get_request_pin_code)) -> HTMLResponse:
        return HTMLResponse(_get_swagger_ui_html(_get_open_json_url(request, r_pin_code), title))

    @doc_pait(pre_depend_list=[_get_request_pin_code])
    def openapi_route(request: Request) -> JSONResponse:
        pait_dict: Dict[str, PaitCoreModel] = load_app(request.app)
        _scheme: str = scheme or request.url.scheme
        url: str = f"{_scheme}://{request.url.hostname}"
        if request.url.port:
            url += f":{request.url.port}"  # pragma: no cover
        pait_openapi: PaitOpenApi = PaitOpenApi(
            pait_dict,
            title=title,
            open_api_server_list=[{"url": url, "description": ""}],
            open_api_tag_list=open_api_tag_list,
        )
        return JSONResponse(pait_openapi.open_api_dict)

    route: Mount = Mount(
        prefix,
        name="api doc",
        routes=[
            Route("/redoc", get_redoc_html, methods=["GET"]),
            Route("/swagger", get_swagger_ui_html, methods=["GET"]),
            Route("/openapi.json", openapi_route, methods=["GET"]),
        ],
    )
    app.routes.append(route)



from typing import Dict, Mapping, Optional

from requests import Response as _Response  # type: ignore
from starlette.testclient import TestClient

from pait.app.base import BaseTestHelper
from pait.model.core import PaitCoreModel

from ._load_app import load_app

__all__ = ["StarletteTestHelper", "TestHelper"]


class TestHelper(BaseTestHelper[_Response]):
    client: TestClient

    def _app_init_field(self) -> None:
        if self.cookie_dict:
            self.header_dict["cookie"] = ";".join([f"{key}={value}" for key, value in self.cookie_dict.items()])

    def _gen_pait_dict(self) -> Dict[str, PaitCoreModel]:
        return load_app(self.client.app)  # type: ignore

    @staticmethod
    def _get_status_code(resp: _Response) -> int:
        return resp.status_code

    @staticmethod
    def _get_content_type(resp: _Response) -> str:
        return resp.headers["content-type"]

    @staticmethod
    def _get_text(resp: _Response) -> str:
        return resp.text

    @staticmethod
    def _get_bytes(resp: _Response) -> bytes:
        return resp.content

    @staticmethod
    def _get_json(resp: _Response) -> dict:
        return resp.json()

    @staticmethod
    def _get_headers(resp: _Response) -> Mapping:
        return resp.headers

    def _replace_path(self, path_str: str) -> Optional[str]:
        if self.path_dict and path_str[0] == "{" and path_str[-1] == "}":
            return self.path_dict[path_str[1:-1]]
        return None

    def _real_request(self, method: str) -> _Response:
        method = method.upper()
        resp: _Response = self.client.request(
            method,
            url=self.path,
            cookies=self.cookie_dict,
            data=self.form_dict,
            json=self.body_dict,
            headers=self.header_dict,
            files=self.file_dict,
        )
        return resp


class StarletteTestHelper(TestHelper):
    """Will remove on version 1.0"""



import json
from types import CodeType
from typing import Dict, List, Type

from pydantic import BaseModel
from pydantic.fields import Undefined

from pait import field
from pait.api_doc.base_parse import FieldDictType, FieldSchemaTypeDict, PaitBaseParse
from pait.model.core import PaitCoreModel
from pait.model.status import PaitStatus
from pait.util import I18n, gen_example_dict_from_schema, join_i18n


class PaitMd(PaitBaseParse):
    """parse pait dict to md doc"""

    def __init__(
        self,
        pait_dict: Dict[str, PaitCoreModel],
        title: str = "Pait Doc",
        use_html_details: bool = True,
        enable_inductive_model: bool = True,
    ):
        """
        :param pait_dict: pait dict
        :param title: Md title
        :param use_html_details: Using HTML syntax-related functions
        :param enable_inductive_model: Whether to enable inductive mode (variable use_html_details must be True)
        """
        if enable_inductive_model and not use_html_details:
            raise ValueError("Not support inductive model when `use_html_details` is False")
        self._use_html_details: bool = use_html_details  # some not support markdown in html
        self._enable_inductive_model: bool = enable_inductive_model
        self._title: str = title

        super().__init__(pait_dict)

        self.content = self.gen_markdown_text()
        self._content_type = ".md"

    @staticmethod
    def gen_md_param_table(field_dict_list: List[FieldSchemaTypeDict], blank_num: int = 8) -> str:
        """
        gen param md table
        :param field_dict_list:
        :param blank_num: table str indent
        :return:
            |param name|type|default value|example|description|other|
            |---|---|---|---|---|---|
            |age|string|10|user age|example value||
        """
        blank_num_str: str = " " * blank_num
        markdown_text: str = (
            f"{blank_num_str}|{join_i18n([I18n.Param, I18n.Name])}|{I18n.Type}|{I18n.Default}"
            f"|{I18n.Example}|{I18n.Desc}|{I18n.Other}|\n"
        )
        markdown_text += f"{blank_num_str}|---|---|---|---|---|---|\n"
        field_dict_list = sorted(field_dict_list, key=lambda x: x["param_name"])
        for field_info_dict in field_dict_list:
            default = field_info_dict["default"]
            if default is Undefined:
                default = f"**`{I18n.Required}`**"
            type_ = field_info_dict["type"]
            if default is Undefined:
                type_ = f"**`{I18n.Required}`**"
            example_value = ""
            if "other" in field_info_dict:
                example_value = field_info_dict["other"].pop("example", "")
            other_value = ", ".join(
                [f"[`{key}:{value}`]" for key, value in (field_info_dict["other"] or dict()).items()]
            )
            markdown_text += (
                f"{blank_num_str}"
                f"|{field_info_dict['param_name'] or ' '}"
                f"|{type_ or ' '}"
                f"|{default or ' '}"
                f"|{example_value or ' '}"
                f"|{field_info_dict['description'] or ' '}"
                f"|{other_value}"
                f"|\n"
            )
        return markdown_text

    def gen_markdown_text(self) -> str:
        markdown_text: str = f"# {self._title}\n"
        for group in self._group_list:
            # group
            if self._use_html_details and self._enable_inductive_model:
                markdown_text += f"<details><summary>{I18n.Group}: {group}</summary>\n\n"
            else:
                markdown_text += f"## {I18n.Group}: {group}\n\n"
            for pait_model in self._group_pait_dict[group]:
                # func info
                if pait_model.status in (
                    PaitStatus.abnormal,
                    PaitStatus.maintenance,
                    PaitStatus.archive,
                    PaitStatus.abandoned,
                ):
                    markdown_text += f"### {I18n.Name}: ~~{pait_model.operation_id}~~\n\n"
                else:
                    markdown_text += f"### {I18n.Name}: {pait_model.operation_id}\n\n"

                if pait_model.desc:
                    markdown_text += f"\n\n**{I18n.Desc}**:{pait_model.desc}\n\n"

                # func or interface details
                func_code: CodeType = pait_model.func.__code__  # type: ignore
                status_text: str = ""
                if pait_model.status in (PaitStatus.test, PaitStatus.design, PaitStatus.dev, PaitStatus.integration):
                    status_text = f"<font color=#00BFFF>{pait_model.status.value}</font>"
                elif pait_model.status in (PaitStatus.release, PaitStatus.complete):
                    status_text = f"<font color=#32CD32>{pait_model.status.value}</font>"
                elif pait_model.status in (
                    PaitStatus.abnormal,
                    PaitStatus.maintenance,
                    PaitStatus.archive,
                    PaitStatus.abandoned,
                ):
                    status_text = f"<font color=#DC143C>{pait_model.status.value}</font>"
                elif pait_model.status:
                    status_text = f"{pait_model.status.value}"
                markdown_text += f"- {join_i18n([I18n.API, I18n.Info])}\n\n"
                markdown_text += f"{' ' * 4}|{I18n.Author}|{I18n.Status}|{I18n.Func}|{I18n.Summary}|\n"
                markdown_text += f"{' ' * 4}|---|---|---|---|\n"
                markdown_text += (
                    f"{' ' * 4}|{','.join(pait_model.author) if pait_model.author else ''}"
                    f"{' ' * 4}|{status_text}"
                    f'{" " * 4}|<abbr title="file:{pait_model.func_path or func_code.co_filename};'
                    f'line: {func_code.co_firstlineno}">'
                    f"{pait_model.func.__qualname__}</abbr>"
                    f"|{' ' * 4}{pait_model.summary}|\n"
                )

                # request info
                markdown_text += f"- {I18n.Path}: {pait_model.path}\n"
                markdown_text += f"- {I18n.Method}: {','.join(pait_model.method_list)}\n"
                markdown_text += f"- {I18n.Request}:\n"

                all_field_dict: FieldDictType = self._parse_pait_model_to_field_dict(pait_model)
                field_class_list: List[Type[field.BaseField]] = sorted(
                    [i for i in all_field_dict.keys()], key=lambda x: x.get_field_name()
                )
                # request body info
                for field_class in field_class_list:
                    if not issubclass(field_class, field.BaseField):
                        continue
                    field_dict_list = all_field_dict[field_class]
                    markdown_text += f"{' ' * 4}- {field_class.get_field_name().capitalize()} {I18n.Param}\n\n"
                    markdown_text += self.gen_md_param_table(field_dict_list)

                # response info
                markdown_text += f"- {I18n.Response}:\n\n"
                if pait_model.response_model_list:
                    for resp_model_class in pait_model.response_model_list:
                        resp_model = resp_model_class()
                        markdown_text += f"{' ' * 4}- {resp_model.name or resp_model.__class__.__name__}\n\n"
                        markdown_text += f"{' ' * 8}- {join_i18n([I18n.Response, I18n.Info])}\n\n"
                        markdown_text += (
                            f"{' ' * 12}|{join_i18n([I18n.Status, I18n.Code])}|"
                            f"{join_i18n([I18n.Media, I18n.Type])}|{I18n.Desc}|\n"
                        )
                        markdown_text += f"{' ' * 12}|---|---|---|\n"
                        markdown_text += (
                            f"{' ' * 12}|{','.join([str(i) for i in resp_model.status_code])}"
                            f"|{resp_model.media_type}"
                            f"|{resp_model.description}"
                            f"|\n"
                        )
                        if resp_model.header:
                            markdown_text += f"{' ' * 8}- Header\n"
                            markdown_text += f"{' ' * 12}{resp_model.header}\n"

                        if isinstance(resp_model.response_data, type) and issubclass(
                            resp_model.response_data, BaseModel
                        ):
                            markdown_text += f"{' ' * 8}- {join_i18n([I18n.Response, I18n.Data])}\n\n"
                            field_dict_list = self._parse_schema(resp_model.response_data.schema())
                            markdown_text += self.gen_md_param_table(field_dict_list, blank_num=12)
                            markdown_text += (
                                f"{' ' * 8}- " f"{join_i18n([I18n.Example, I18n.Response, 'Json', I18n.Data])} \n\n"
                            )
                            example_dict: dict = gen_example_dict_from_schema(resp_model.response_data.schema())
                            blank_num_str: str = " " * 12
                            json_str: str = "\n".join(
                                [blank_num_str + i for i in json.dumps(example_dict, indent=2).split("\n")]
                            )
                            markdown_text += f"{blank_num_str}```json\n{json_str}\n{blank_num_str}```\n\n"
                markdown_text += "\n"
            if self._use_html_details:
                markdown_text += "</details>"
        return markdown_text



from typing import Any

from pait.data import PaitData
from pait.model import tag
from pait.model.config import Config
from pait.model.util import sync_config_data_to_pait_core_model

__all__ = ["config", "pait_data"]

# In order to reduce the intrusion of pait to the application framework,
# pait conducts data interaction through PaitData and Config
pait_data: PaitData = PaitData()
config: Config = Config()

# pait_data and config cannot refer to each other within the module
# By replacing the init_config function, the user can change the pait_data data when calling config.init_config
_real_config_init_config_method = config.init_config


def _after_config_init(*args: Any, **kwargs: Any) -> None:
    _real_config_init_config_method(*args, **kwargs)
    for app_name, real_pait_id_dict in pait_data.pait_id_dict.items():
        for pait_id, pait_info_model in real_pait_id_dict.items():
            sync_config_data_to_pait_core_model(config, pait_info_model)


setattr(config, config.init_config.__name__, _after_config_init)
config.tag_dict = getattr(tag, "_tag_dict", {})



from typing import Any, List

from pait.exceptions import CheckValueError
from pait.plugin.base import BaseAsyncPlugin, BasePlugin, PluginProtocol
from pait.util import gen_tip_exc

__all__ = ["AsyncAtMostOneOfPlugin", "AtMostOneOfPlugin"]


class AtMostOneOfPluginProtocol(PluginProtocol):
    """
    Check whether each group of parameters appear at the same time
    """

    is_pre_core: bool = False

    def __init__(self, *, at_most_one_of_list: List[List[str]]):
        super().__init__()

        self.at_most_one_of_list: List[List[str]] = at_most_one_of_list

    def check_param(self, *args: Any, **kwargs: Any) -> None:
        try:
            for at_most_one_of in self.at_most_one_of_list:
                if len([i for i in at_most_one_of if kwargs.get(i, None) is not None]) > 1:
                    raise CheckValueError(f"requires at most one of param {' or '.join(at_most_one_of)}")
        except Exception as e:
            raise e from gen_tip_exc(self.pait_core_model.func, e)


class AtMostOneOfPlugin(AtMostOneOfPluginProtocol, BasePlugin):
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        self.check_param(*args, **kwargs)
        return self.call_next(*args, **kwargs)


class AsyncAtMostOneOfPlugin(AtMostOneOfPluginProtocol, BaseAsyncPlugin):
    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        self.check_param(*args, **kwargs)
        return await self.call_next(*args, **kwargs)



import inspect
import logging
from functools import wraps
from typing import Any, Callable, List, Optional, Tuple, Type, TypeVar, Union

from pydantic import BaseConfig

from pait.app.base import BaseAppHelper
from pait.g import config, pait_data
from pait.model.core import PaitCoreModel
from pait.model.response import PaitBaseResponseModel
from pait.model.status import PaitStatus
from pait.model.tag import Tag
from pait.model.util import sync_config_data_to_pait_core_model
from pait.plugin.base import BaseAsyncPlugin, BasePlugin, PluginManager
from pait.util import get_func_sig

_AppendT = TypeVar("_AppendT", list, tuple)
_PaitT = TypeVar("_PaitT", bound="Pait")
_PluginT = Union[BasePlugin, BaseAsyncPlugin]
TagT = Union[str, Tag]


class Pait(object):
    app_helper_class: "Type[BaseAppHelper]"

    def __init__(
        self,
        pydantic_model_config: Optional[Type[BaseConfig]] = None,
        # param check
        pre_depend_list: Optional[List[Callable]] = None,
        # doc
        author: Optional[Tuple[str, ...]] = None,
        desc: Optional[str] = None,
        summary: Optional[str] = None,
        name: Optional[str] = None,
        status: Optional[PaitStatus] = None,
        group: Optional[str] = None,
        tag: Optional[Tuple[TagT, ...]] = None,
        response_model_list: Optional[List[Type[PaitBaseResponseModel]]] = None,
        plugin_list: Optional[List[PluginManager]] = None,
        post_plugin_list: Optional[List[PluginManager]] = None,
    ):

        check_cls_param_list: List[str] = ["app_helper_class"]
        for cls_param in check_cls_param_list:
            if not getattr(self, cls_param, None):
                raise ValueError(
                    f"Please specify the value of the {cls_param} parameter, you can refer to `pait.app.xxx`"
                )
        if not isinstance(self.app_helper_class, type):
            raise TypeError(f"{self.app_helper_class} must be class")
        if not issubclass(self.app_helper_class, BaseAppHelper):
            raise TypeError(f"{self.app_helper_class} must sub from {BaseAppHelper.__class__.__name__}")

        self._pydantic_model_config: Optional[Type[BaseConfig]] = pydantic_model_config
        # param check
        self._pre_depend_list: Optional[List[Callable]] = pre_depend_list
        # doc
        self._author: Optional[Tuple[str, ...]] = author
        self._desc: Optional[str] = desc
        self._summary: Optional[str] = summary
        self._name: Optional[str] = name
        self._status: Optional[PaitStatus] = status
        self._group: Optional[str] = group
        self._tag: Optional[Tuple[TagT, ...]] = tag
        self._response_model_list: Optional[List[Type[PaitBaseResponseModel]]] = response_model_list
        self._plugin_list: Optional[List[PluginManager]] = plugin_list
        self._post_plugin_list: Optional[List[PluginManager]] = post_plugin_list

    @staticmethod
    def _append_data(
        target_container: Optional[_AppendT],
        append_container: Optional[_AppendT],
        self_container: Optional[_AppendT],
    ) -> Optional[_AppendT]:
        if append_container:
            return append_container + (self_container or append_container.__class__())
        else:
            return target_container or self_container

    def create_sub_pait(
        self: _PaitT,
        pydantic_model_config: Optional[Type[BaseConfig]] = None,
        # param check
        pre_depend_list: Optional[List[Callable]] = None,
        append_pre_depend_list: Optional[List[Callable]] = None,
        # doc
        author: Optional[Tuple[str, ...]] = None,
        append_author: Optional[Tuple[str, ...]] = None,
        desc: Optional[str] = None,
        summary: Optional[str] = None,
        name: Optional[str] = None,
        status: Optional[PaitStatus] = None,
        group: Optional[str] = None,
        tag: Optional[Tuple[TagT, ...]] = None,
        append_tag: Optional[Tuple[TagT, ...]] = None,
        response_model_list: Optional[List[Type[PaitBaseResponseModel]]] = None,
        append_response_model_list: Optional[List[Type[PaitBaseResponseModel]]] = None,
        plugin_list: Optional[List[PluginManager]] = None,
        append_plugin_list: Optional[List[PluginManager]] = None,
        post_plugin_list: Optional[List[PluginManager]] = None,
        append_post_plugin_list: Optional[List[PluginManager]] = None,
    ) -> _PaitT:
        pre_depend_list = self._append_data(pre_depend_list, append_pre_depend_list, self._pre_depend_list)
        author = self._append_data(author, append_author, self._author)
        tag = self._append_data(tag, append_tag, self._tag)
        response_model_list = self._append_data(
            response_model_list, append_response_model_list, self._response_model_list
        )
        plugin_list = self._append_data(plugin_list, append_plugin_list, self._plugin_list)
        post_plugin_list = self._append_data(post_plugin_list, append_post_plugin_list, self._post_plugin_list)

        return self.__class__(
            pydantic_model_config=pydantic_model_config or self._pydantic_model_config,
            desc=desc or self._desc,
            summary=summary or self._summary,
            name=name or self._name,
            status=status or self._status,
            pre_depend_list=pre_depend_list,
            author=author,
            group=group,
            tag=tag,
            response_model_list=response_model_list,
            plugin_list=plugin_list,
            post_plugin_list=post_plugin_list,
        )

    @staticmethod
    def _plugin_manager_handler(
        pait_core_model: PaitCoreModel,
        args: Any,
        kwargs: Any,
        func: Callable,
    ) -> List[_PluginT]:
        plugin_instance_list: List[_PluginT] = []

        pre_plugin: Callable = func
        for plugin_manager in pait_core_model.plugin_list:
            plugin_instance: _PluginT = plugin_manager.get_plugin()
            plugin_instance.__post_init__(pait_core_model, args, kwargs)
            plugin_instance.call_next = pre_plugin  # type: ignore
            pre_plugin = plugin_instance
            plugin_instance_list.append(plugin_instance)
        plugin_instance_list.reverse()
        return plugin_instance_list

    def __call__(
        self,
        pydantic_model_config: Optional[Type[BaseConfig]] = None,
        # param check
        pre_depend_list: Optional[List[Callable]] = None,
        append_pre_depend_list: Optional[List[Callable]] = None,
        # doc
        author: Optional[Tuple[str, ...]] = None,
        append_author: Optional[Tuple[str, ...]] = None,
        desc: Optional[str] = None,
        summary: Optional[str] = None,
        name: Optional[str] = None,
        status: Optional[PaitStatus] = None,
        group: Optional[str] = None,
        tag: Optional[Tuple[TagT, ...]] = None,
        append_tag: Optional[Tuple[TagT, ...]] = None,
        response_model_list: Optional[List[Type[PaitBaseResponseModel]]] = None,
        append_response_model_list: Optional[List[Type[PaitBaseResponseModel]]] = None,
        plugin_list: Optional[List[PluginManager]] = None,
        append_plugin_list: Optional[List[PluginManager]] = None,
        post_plugin_list: Optional[List[PluginManager]] = None,
        append_post_plugin_list: Optional[List[PluginManager]] = None,
    ) -> Callable:
        app_name: str = self.app_helper_class.app_name
        pydantic_model_config = pydantic_model_config or self._pydantic_model_config
        desc = desc or self._desc
        summary = summary or self._summary
        name = name or self._name
        group = group or self._group

        pre_depend_list = self._append_data(pre_depend_list, append_pre_depend_list, self._pre_depend_list)
        author = self._append_data(author, append_author, self._author)
        tag = self._append_data(tag, append_tag, self._tag)
        response_model_list = self._append_data(
            response_model_list, append_response_model_list, self._response_model_list
        )
        plugin_list = self._append_data(plugin_list, append_plugin_list, self._plugin_list)
        post_plugin_list = self._append_data(post_plugin_list, append_post_plugin_list, self._post_plugin_list)

        ###############
        # tag handler #
        ###############
        new_tag: List[str] = []
        if tag:
            for _tag in tag:
                if isinstance(_tag, Tag):
                    _tag = _tag.name
                else:
                    logging.warning(
                        "In later versions tag only supports Tag class, and does not support str type"
                    )  # pragma: no cover
                new_tag.append(_tag)

        def wrapper(func: Callable) -> Callable:
            # Pre-parsing function signatures
            get_func_sig(func)
            # gen pait core model and register to pait data
            pait_core_model: PaitCoreModel = PaitCoreModel(
                author=author,
                app_helper_class=self.app_helper_class,
                desc=desc,
                summary=summary,
                func=func,
                func_name=name,
                status=status,
                group=group,
                tag=tuple(new_tag),
                response_model_list=response_model_list,
                pre_depend_list=pre_depend_list,
                pydantic_model_config=pydantic_model_config,
                plugin_list=plugin_list,
                post_plugin_list=post_plugin_list,
            )
            sync_config_data_to_pait_core_model(config, pait_core_model)
            pait_data.register(app_name, pait_core_model)
            if inspect.iscoroutinefunction(func):

                @wraps(func)
                async def dispatch(*args: Any, **kwargs: Any) -> Callable:
                    first_plugin: BaseAsyncPlugin = self._plugin_manager_handler(  # type: ignore
                        pait_core_model, args, kwargs, func
                    )[0]
                    return await first_plugin(*args, **kwargs)

                return dispatch
            else:

                @wraps(func)
                def dispatch(*args: Any, **kwargs: Any) -> Callable:
                    first_plugin: BasePlugin = self._plugin_manager_handler(  # type: ignore
                        pait_core_model, args, kwargs, func
                    )[0]
                    return first_plugin(*args, **kwargs)

                return dispatch

        return wrapper



from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from json import JSONEncoder
from typing import Any, Dict, List, Optional, Set, Tuple, Type

from pydantic import BaseConfig

from pait.model.response import PaitBaseResponseModel
from pait.model.status import PaitStatus
from pait.util import I18nTypedDict, change_local
from pait.util import i18n_config_dict as pait_i18n_config_dict
from pait.util import i18n_local as pait_i18n_local
from pait.util import json_type_default_value_dict as pait_json_type_default_value_dict
from pait.util import python_type_default_value_dict as pait_python_type_default_value_dict


class CustomJSONEncoder(JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, datetime):
            return int(obj.timestamp())
        elif isinstance(obj, date):
            return obj.strftime("%Y-%m-%d")
        elif isinstance(obj, Decimal):
            return float(obj)
        elif isinstance(obj, Enum):
            return obj.value
        else:
            return super().default(obj)


class Config(object):
    """
    Provide pait parameter configuration, the init_config method is valid only before the application service is started
    """

    __initialized: bool = False

    def __init__(self) -> None:
        self.author: Tuple[str, ...] = ("",)
        self.status: PaitStatus = PaitStatus.undefined
        self.block_http_method_set: Set[str] = set()
        self.default_response_model_list: List[Type[PaitBaseResponseModel]] = []
        self.json_encoder: Type[JSONEncoder] = CustomJSONEncoder
        self.default_pydantic_model_config: Type[BaseConfig] = BaseConfig
        self.tag_dict: Dict[str, str] = {}
        self.i18n_local: str = pait_i18n_local

    def __setattr__(self, key: Any, value: Any) -> None:
        if not self.__initialized:
            super().__setattr__(key, value)
        else:
            raise RuntimeError("Can not set new value in runtime")

    def init_config(
        self,
        author: Optional[Tuple[str, ...]] = None,
        status: Optional[PaitStatus] = None,
        default_response_model_list: Optional[List[Type[PaitBaseResponseModel]]] = None,
        block_http_method_set: Optional[Set[str]] = None,
        json_type_default_value_dict: Optional[Dict[str, Any]] = None,
        python_type_default_value_dict: Optional[Dict[type, Any]] = None,
        json_encoder: Optional[Type[JSONEncoder]] = None,
        default_pydantic_model_config: Optional[Type[BaseConfig]] = None,
        i18n_local: str = pait_i18n_local,
        i18n_config_dict: Optional[Dict[str, I18nTypedDict]] = None,
    ) -> None:
        """
        :param author:  Only @pait(author=None) will be called to change the configuration
        :param status:  Only @pait(status=None) will be called to change the configuration
        :param default_response_model_list: Add a default response structure for all routing handles
        :param block_http_method_set:
            Under normal circumstances, pait.load_app can obtain the http method of the routing handle.
            However, some application frameworks such as flask will automatically add optional http methods
             to the handle, which is great, but you may not want to use them in pait, and pait will not
             automatically recognize them, so you can use parameters to disable certain http method
        :param json_type_default_value_dict: Configure default values for each json type
        :param python_type_default_value_dict: Configure default values for each python type
        :param json_encoder: Define certain types of serialization rules
        :param default_pydantic_model_config: pait route gen pydantic model default config
        :return:
        """
        if author:
            self.author = author
        if status:
            self.status = status
        if default_response_model_list:
            self.default_response_model_list = default_response_model_list
        if block_http_method_set:
            self.block_http_method_set = block_http_method_set
        if json_type_default_value_dict:
            pait_json_type_default_value_dict.update(json_type_default_value_dict)
        if python_type_default_value_dict:
            pait_python_type_default_value_dict.update(python_type_default_value_dict)
        if json_encoder:
            self.json_encoder = json_encoder
        if default_pydantic_model_config:
            self.default_pydantic_model_config = default_pydantic_model_config

        if pait_i18n_local:
            change_local(pait_i18n_local)
        if i18n_config_dict:
            pait_i18n_config_dict.update(i18n_config_dict)

        self.__initialized = True

    @property
    def initialized(self) -> bool:
        """Judge whether it has been initialized, it can only be initialized once per run"""
        return self.__initialized



import re
from typing import Any, Dict, List

from typing_extensions import TypedDict

I18nTypedDict = TypedDict("I18nTypedDict", {"en": str, "zh-cn": str})

i18n_config_dict: Dict[str, I18nTypedDict] = {
    "Group": {
        "en": "Group",
        "zh-cn": "组",
    },
    "Name": {
        "en": "Name",
        "zh-cn": "名称",
    },
    "Desc": {
        "en": "Desc",
        "zh-cn": "描述",
    },
    "API": {
        "en": "API",
        "zh-cn": "API",
    },
    "Info": {
        "en": "Info",
        "zh-cn": "信息",
    },
    "Author": {
        "en": "Author",
        "zh-cn": "作者",
    },
    "Status": {
        "en": "Status",
        "zh-cn": "状态",
    },
    "Func": {
        "en": "Func",
        "zh-cn": "函数",
    },
    "Summary": {
        "en": "Summary",
        "zh-cn": "摘要",
    },
    "Path": {"en": "Path", "zh-cn": "路径"},
    "Method": {"en": "Method", "zh-cn": "方法"},
    "Request": {"en": "Request", "zh-cn": "请求"},
    "Response": {"en": "Response", "zh-cn": "响应"},
    "Code": {"en": "Code", "zh-cn": "码"},
    "Media": {"en": "Media", "zh-cn": "媒体"},
    "Data": {"en": "Data", "zh-cn": "数据"},
    "Example": {"en": "Example", "zh-cn": "示例"},
    "Default": {"en": "Default", "zh-cn": "默认"},
    "Other": {"en": "Other", "zh-cn": "其它"},
    "Required": {"en": "Required", "zh-cn": "必填"},
    "Type": {"en": "Type", "zh-cn": "类型"},
    "Param": {"en": "Param", "zh-cn": "参数"},
}
i18n_local: str = "en"


class I18nHelper(object):
    def __init__(self, name: str):
        # in python 3.10,  can use __set_name__
        self.name: str = name

    def __get__(self, instance: Any, owner: Any) -> Any:
        default: str = " ".join((re.sub(r"(?P<key>[A-Z])", r"_\g<key>", self.__class__.__name__)).split("_"))
        return i18n_config_dict[self.name].get(i18n_local, default)

    @classmethod
    def i(cls, name: str) -> Any:
        return cls(name)


class I18n(object):
    Group: str = I18nHelper.i("Group")
    Name: str = I18nHelper.i("Name")
    Desc: str = I18nHelper.i("Desc")
    API: str = I18nHelper.i("API")
    Info: str = I18nHelper.i("Info")
    Author: str = I18nHelper.i("Author")
    Status: str = I18nHelper.i("Status")
    Func: str = I18nHelper.i("Func")
    Summary: str = I18nHelper.i("Summary")
    Path: str = I18nHelper.i("Path")
    Method: str = I18nHelper.i("Method")
    Request: str = I18nHelper.i("Request")
    Response: str = I18nHelper.i("Response")
    Code: str = I18nHelper.i("Code")
    Media: str = I18nHelper.i("Media")
    Data: str = I18nHelper.i("Data")
    Example: str = I18nHelper.i("Example")
    Default: str = I18nHelper.i("Default")
    Other: str = I18nHelper.i("Other")
    Required: str = I18nHelper.i("Required")
    Type: str = I18nHelper.i("Type")
    Param: str = I18nHelper.i("Param")


def change_local(lang: str) -> None:
    global i18n_local

    if lang not in ("en", "zh-cn"):
        if not lang.startswith("customer_"):
            raise ValueError(f"Not support {lang}")
    i18n_local = lang


def join_i18n(i18n_list: List[str]) -> str:
    if i18n_local == "zh-cn":
        return "".join(i18n_list)
    else:
        return " ".join(i18n_list)



class PaitBaseException(Exception):
    pass


class CheckValueError(PaitBaseException):
    pass


class ParseTypeError(PaitBaseException):
    pass


class PaitBaseParamException(PaitBaseException):
    def __init__(self, param: str, msg: str):
        super().__init__(msg)
        self.param: str = param
        self.msg: str = msg


class NotFoundFieldException(PaitBaseParamException):
    pass


class NotFoundValueException(PaitBaseParamException):
    pass


class FieldValueTypeException(PaitBaseParamException):
    pass


class TipException(PaitBaseException):
    def __init__(self, msg: str, exc: Exception):
        super().__init__(msg)
        self.exc: Exception = exc
